#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2000-2006  Donald N. Allingham
# Copyright (C) 2010       Nick Hall
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

__all__ = [
    "MonitoredCheckbox",
    "MonitoredEntry",
    "MonitoredEntryIndicator",
    "MonitoredSpinButton",
    "MonitoredText",
    "MonitoredType",
    "MonitoredDataType",
    "MonitoredMenu",
    "MonitoredStrMenu",
    "MonitoredDate",
    "MonitoredComboSelectedEntry",
    "MonitoredTagList",
]

# -------------------------------------------------------------------------
#
# Standard python modules
#
# -------------------------------------------------------------------------
import logging

_LOG = logging.getLogger(".widgets.monitoredwidgets")

# -------------------------------------------------------------------------
#
# GTK/Gnome modules
#
# -------------------------------------------------------------------------
from gi.repository import GObject
from gi.repository import Gtk
from gi.repository import Gdk
from gi.repository import Pango

# -------------------------------------------------------------------------
#
# Gramps modules
#
# -------------------------------------------------------------------------
from gramps.gen.const import GRAMPS_LOCALE as glocale

_ = glocale.translation.gettext
from ..autocomp import StandardCustomSelector, fill_entry
from gramps.gen.datehandler import displayer, parser
from gramps.gen.lib.date import Date, NextYear
from gramps.gen.errors import ValidationError

# -------------------------------------------------------------------------
#
# constants
#
# ------------------------------------------------------------------------

_RETURN = Gdk.keyval_from_name("Return")
_KP_ENTER = Gdk.keyval_from_name("KP_Enter")


# -------------------------------------------------------------------------
#
# MonitoredCheckbox class
#
# -------------------------------------------------------------------------
class MonitoredCheckbox:
    def __init__(self, obj, button, set_val, get_val, on_toggle=None, readonly=False):
        self.button = button
        self.button.connect("toggled", self._on_toggle)
        self.on_toggle = on_toggle
        self.obj = obj
        self.set_val = set_val
        self.get_val = get_val
        self.button.set_active(get_val())
        self.button.set_sensitive(not readonly)

    def _on_toggle(self, obj):
        self.set_val(obj.get_active())
        if self.on_toggle:
            self.on_toggle(self.get_val())


##    def destroy(self):
##        """
##        Unset all elements that can prevent garbage collection
##        """
##        self.set_val = None
##        self.get_val = None
##        self.obj = None


# -------------------------------------------------------------------------
#
# MonitoredEntry class
#
# -------------------------------------------------------------------------
class MonitoredEntry:
    def __init__(
        self, obj, set_val, get_val, read_only=False, autolist=None, changed=None
    ):
        self.obj = obj
        self.set_val = set_val
        self.get_val = get_val
        self.changed = changed

        if get_val():
            self.obj.set_text(get_val())
        self.obj.connect("changed", self._on_change)
        self.obj.connect("focus-out-event", self._on_quit)
        self.obj.set_editable(not read_only)

        if autolist:
            fill_entry(obj, autolist)

    ##    def destroy(self):
    ##        """
    ##        Unset all elements that can prevent garbage collection
    ##        """
    ##        self.set_val = None
    ##        self.get_val = None
    ##        self.obj = None

    def reinit(self, set_val, get_val):
        self.set_val = set_val
        self.get_val = get_val
        self.update()

    def set_text(self, text):
        self.obj.set_text(text)

    def connect(self, signal, callback, *data):
        self.obj.connect(signal, callback, *data)

    def _on_quit(self, obj, event):
        self.set_val(obj.get_text().strip())

    def _on_change(self, obj):
        self.set_val(obj.get_text())
        if self.changed:
            self.changed(obj)

    def force_value(self, value):
        self.obj.set_text(value)

    def get_value(self):
        return str(self.obj.get_text())

    def enable(self, value):
        self.obj.set_sensitive(value)
        self.obj.set_editable(value)

    def grab_focus(self):
        self.obj.grab_focus()

    def update(self):
        if self.get_val() is not None:
            self.obj.set_text(self.get_val())


# -------------------------------------------------------------------------
#
# MonitoredEntryIndicator class
#
# -------------------------------------------------------------------------
class MonitoredEntryIndicator(MonitoredEntry):
    """
    Show an Entry box with an indicator in it that disappears when
    entry becomes active
    """

    def __init__(
        self,
        obj,
        set_val,
        get_val,
        indicator,
        read_only=False,
        autolist=None,
        changed=None,
    ):
        MonitoredEntry.__init__(
            self, obj, set_val, get_val, read_only, autolist, changed
        )
        self.obj.set_placeholder_text(indicator)


# -------------------------------------------------------------------------
#
# MonitoredSpinButton class
#
# -------------------------------------------------------------------------
class MonitoredSpinButton:
    """
    Class for signal handling of spinbuttons.
    (Code is a modified copy of :class:`MonitoredEntry`)
    """

    def __init__(
        self, obj, set_val, get_val, read_only=False, autolist=None, changed=None
    ):
        """
        :param obj: widget to be monitored
        :type obj: Gtk.SpinButton
        :param set_val: callback to be called when obj is changed
        :param get_val: callback to be called to retrieve value for obj
        :param read_only: If SpinButton is read only.
        """

        self.obj = obj
        self.set_val = set_val
        self.get_val = get_val
        self.changed = changed

        if get_val():
            self.obj.set_value(get_val())
        self.obj.connect("value-changed", self._on_change)
        self.obj.set_editable(not read_only)

        if autolist:
            fill_entry(obj, autolist)

    ##    def destroy(self):
    ##        """
    ##        Unset all elements that can prevent garbage collection
    ##        """
    ##        self.set_val = None
    ##        self.get_val = None
    ##        self.obj = None

    def reinit(self, set_val, get_val):
        """
        Reinitialize class with the specified callback functions.

        :param set_val: callback to be called when SpinButton is changed
        :param get_val: callback to be called to retrieve value for SpinButton
        """

        self.set_val = set_val
        self.get_val = get_val
        self.update()

    def set_value(self, value):
        """
        Set the value of the monitored widget to the specified value.

        :param value: Value to be set.
        """

        self.obj.set_value(value)

    def connect(self, signal, callback):
        """
        Connect the signal of monitored widget to the specified callback.

        :param signal: Signal prototype for which a connection should be set up.
        :param callback: Callback function to be called when signal is emitted.
        """

        self.obj.connect(signal, callback)

    def _on_change(self, obj):
        """
        Event handler to be called when the monitored widget is changed.

        :param obj: Widget that has been changed.
        :type obj: Gtk.SpinButton
        """

        self.set_val(obj.get_value())
        if self.changed:
            self.changed(obj)

    def force_value(self, value):
        """
        Set the value of the monitored widget to the specified value.

        :param value: Value to be set.
        """

        self.obj.set_value(value)

    def get_value(self):
        """
        Get the current value of the monitored widget.

        :returns: Current value of monitored widget.
        """

        return self.obj.get_value()

    def enable(self, value):
        """
        Change the property editable and sensitive of the monitored widget to
        value.

        :param value: If widget should be editable or deactivated.
        :type value: bool
        """

        self.obj.set_sensitive(value)
        self.obj.set_editable(value)

    def grab_focus(self):
        """
        Assign the keyboard focus to the monitored widget.
        """

        self.obj.grab_focus()

    def update(self):
        """
        Updates value of monitored SpinButton with the value returned by the
        get_val callback.
        """

        if self.get_val():
            self.obj.set_value(self.get_val())


# -------------------------------------------------------------------------
#
# MonitoredText class
#
# -------------------------------------------------------------------------
class MonitoredText:
    def __init__(self, obj, set_val, get_val, read_only=False):
        self.buf = obj.get_buffer()
        self.set_val = set_val
        self.get_val = get_val

        if get_val():
            self.buf.set_text(get_val())
        self.buf.connect("changed", self.on_change)
        obj.set_editable(not read_only)

    ##    def destroy(self):
    ##        """
    ##        Unset all elements that can prevent garbage collection
    ##        """
    ##        self.set_val = None
    ##        self.get_val = None
    ##        self.buf = None

    def on_change(self, obj):
        s, e = self.buf.get_bounds()
        self.set_val(str(self.buf.get_text(s, e, False)))


# -------------------------------------------------------------------------
#
# MonitoredType class
#
# -------------------------------------------------------------------------
class MonitoredType:
    def __init__(
        self, obj, set_val, get_val, mapping, custom, readonly=False, custom_values=None
    ):
        self.set_val = set_val
        self.get_val = get_val

        self.obj = obj

        val = get_val()
        if val:
            default = val[0]
        else:
            default = None

        self.sel = StandardCustomSelector(
            mapping, obj, custom, default, additional=custom_values
        )

        self.set_val(self.sel.get_values())
        self.obj.set_sensitive(not readonly)
        self.obj.connect("changed", self.on_change)

    ##    def destroy(self):
    ##        """
    ##        Unset all elements that can prevent garbage collection
    ##        """
    ##        self.set_val = None
    ##        self.get_val = None
    ##        self.obj = None

    def reinit(self, set_val, get_val):
        self.set_val = set_val
        self.get_val = get_val
        self.update()

    def update(self):
        if self.get_val():
            self.sel.set_values(self.get_val())

    def on_change(self, obj):
        self.set_val(self.sel.get_values())


# -------------------------------------------------------------------------
#
# MonitoredDataType class
#
# -------------------------------------------------------------------------
class MonitoredDataType:
    def __init__(
        self,
        obj,
        set_val,
        get_val,
        readonly=False,
        custom_values=None,
        ignore_values=None,
    ):
        """
        Constructor for the MonitoredDataType class.

        :param obj: Existing ComboBox widget to use with has_entry=True.
        :type obj: Gtk.ComboBox
        :param set_val: The function that sets value of the type in the object
        :type set_val: method
        :param get_val: The function that gets value of the type in the object.
                        This returns a GrampsType, of which get_map returns all
                        possible types
        :type get_val: method
        :param custom_values: Extra values to show in the combobox. These can be
                              text of custom type, tuple with type info or
                              GrampsType class
        :type custom_values: list of str, tuple or GrampsType
        :param ignore_values: list of values not to show in the combobox. If the
                              result of get_val is in these, it is not ignored
        :type ignore_values: list of int
        """
        self.set_val = set_val
        self.get_val = get_val

        self.obj = obj

        val = get_val()

        if val:
            default = int(val)
        else:
            default = None

        map = get_val().get_map().copy()
        if ignore_values:
            for key in list(map.keys()):
                if key in ignore_values and key not in (None, default):
                    del map[key]

        self.sel = StandardCustomSelector(
            map,
            obj,
            get_val().get_custom(),
            default,
            additional=custom_values,
            menu=get_val().get_menu(),
        )

        self.sel.set_values((int(get_val()), str(get_val())))
        self.obj.set_sensitive(not readonly)
        self.obj.connect("changed", self.on_change)

    ##    def destroy(self):
    ##        """
    ##        Unset all elements that can prevent garbage collection
    ##        """
    ##        self.set_val = None
    ##        self.get_val = None
    ##        self.obj = None

    def reinit(self, set_val, get_val):
        self.set_val = set_val
        self.get_val = get_val
        self.update()

    def fix_value(self, value):
        if value[0] == self.get_val().get_custom():
            return value
        else:
            return (value[0], "")

    def update(self):
        val = self.get_val()
        if isinstance(val, tuple):
            self.sel.set_values(val)
        else:
            self.sel.set_values((int(val), str(val)))

    def on_change(self, obj):
        value = self.fix_value(self.sel.get_values())
        self.set_val(value)


# -------------------------------------------------------------------------
#
# MonitoredMenu class
#
# -------------------------------------------------------------------------
class MonitoredMenu:
    def __init__(self, obj, set_val, get_val, mapping, readonly=False, changed=None):
        self.set_val = set_val
        self.get_val = get_val

        self.changed = changed
        self.obj = obj

        self.change_menu(mapping)
        self.obj.connect("changed", self.on_change)
        self.obj.set_sensitive(not readonly)

    ##    def destroy(self):
    ##        """
    ##        Unset all elements that can prevent garbage collection
    ##        """
    ##        self.set_val = None
    ##        self.get_val = None
    ##        self.obj = None

    def force(self, value):
        self.obj.set_active(value)

    def change_menu(self, mapping):
        self.data = {}
        self.model = Gtk.ListStore(GObject.TYPE_STRING, GObject.TYPE_INT)
        index = 0
        for t, v in mapping:
            self.model.append(row=[t, v])
            self.data[v] = index
            index += 1
        self.obj.set_model(self.model)
        self.obj.set_active(self.data.get(self.get_val(), 0))

    def on_change(self, obj):
        self.set_val(self.model.get_value(obj.get_active_iter(), 1))
        if self.changed:
            self.changed()


# -------------------------------------------------------------------------
#
# MonitoredStrMenu class
#
# -------------------------------------------------------------------------
class MonitoredStrMenu:
    def __init__(self, obj, set_val, get_val, mapping, readonly=False):
        self.set_val = set_val
        self.get_val = get_val

        self.obj = obj
        self.model = Gtk.ListStore(GObject.TYPE_STRING)

        # Make sure that the menu is visible on small screen devices.
        # Some LDS temples were not visible on a 4 or 5 column layout.
        # See bug #7333
        if len(mapping) > 20:
            self.obj.set_wrap_width(3)

        self.model.append(row=[""])
        index = 0
        self.data = [""]

        default = get_val()
        active = 0

        for t, v in mapping:
            self.model.append(row=[v])
            self.data.append(t)
            index += 1
            if t == default:
                active = index

        self.obj.set_model(self.model)
        self.obj.set_active(active)
        self.obj.connect("changed", self.on_change)
        self.obj.set_sensitive(not readonly)

    ##    def destroy(self):
    ##        """
    ##        Unset all elements that can prevent garbage collection
    ##        """
    ##        self.set_val = None
    ##        self.get_val = None
    ##        self.obj = None
    ##        self.model = None

    def on_change(self, obj):
        self.set_val(self.data[obj.get_active()])


# -------------------------------------------------------------------------
#
# MonitoredDate class
#
# -------------------------------------------------------------------------
class MonitoredDate:
    """
    Class that associates a pixmap with a text widget, providing visual
    feedback that indicates if the text widget contains a valid date.
    """

    def __init__(self, field, button, value, uistate, track, readonly=False):
        """
        Create a connection between the date_obj, text_obj and the pixmap_obj.
        Assigns callbacks to parse and change date when the text
        in text_obj is changed, and to invoke Date Editor when the LED
        button_obj is pressed.
        """
        self.uistate = uistate
        self.track = track
        self.date_obj = value
        self.text_obj = field
        self.button_obj = button

        image = Gtk.Image()
        image.set_from_icon_name("gramps-date-edit", Gtk.IconSize.BUTTON)
        self.button_obj.set_image(image)
        self.button_obj.set_relief(Gtk.ReliefStyle.NORMAL)
        self.pixmap_obj = self.button_obj.get_child()

        self.text_obj.connect("validate", self.validate)
        self.text_obj.connect("content-changed", self.set_date)
        self.button_obj.connect("clicked", self.invoke_date_editor)

        self.text_obj.set_text(displayer.display(self.date_obj))
        self.text_obj.validate()

        self.text_obj.set_editable(not readonly)
        self.button_obj.set_sensitive(not readonly)

    def set_date(self, widget):
        """
        Parse date from text entry to date object
        """
        date = parser.parse(str(self.text_obj.get_text()))
        self.date_obj.copy(date)

    def validate(self, widget, data):
        """
        Validate current date in text entry
        """
        # if text could not be parsed it is assumed invalid
        if self.date_obj.get_modifier() == Date.MOD_TEXTONLY:
            return ValidationError(_("Bad Date"))
        elif self.date_obj.to_calendar(calendar_name=Date.CAL_GREGORIAN) >> NextYear():
            return ValidationError(_("Date more than one year in the future"))

    def invoke_date_editor(self, obj):
        """
        Invokes Date Editor dialog when the user clicks the Calendar button.
        If date was in fact built, sets the date_obj to the newly built
        date.
        """
        from ..editors import EditDate

        date_dialog = EditDate(self.date_obj, self.uistate, self.track)
        the_date = date_dialog.return_date
        self.update_after_editor(the_date)

    def update_after_editor(self, date_obj):
        """
        Update text entry and validate it
        """
        if date_obj:
            # first we set the text entry, that emits 'content-changed'
            # signal thus the date object gets updated too
            self.text_obj.set_text(displayer.display(date_obj))
            self.text_obj.validate()


# -------------------------------------------------------------------------
#
# MonitoredComboSelectedEntry class
#
# -------------------------------------------------------------------------
class MonitoredComboSelectedEntry:
    """
    A MonitoredEntry driven by a Combobox to select what the entry field
    works upon
    """

    def __init__(
        self,
        objcombo,
        objentry,
        textlist,
        set_val_list,
        get_val_list,
        default=0,
        read_only=False,
    ):
        """
        Create a MonitoredComboSelectedEntry
        Objcombo and objentry should be the gtk widgets to use
        textlist is the values that must be used in the combobox
        Every value needs an entry in set/get_val_list with the data retrieval
        and storage method of the data entered in the entry box
        Read_only should be true if no changes may be done
        default is the entry in the combobox that must be preselected
        """
        self.objcombo = objcombo
        self.objentry = objentry
        self.set_val_list = set_val_list
        self.get_val_list = get_val_list

        # fill the combobox, set on a specific entry
        self.mapping = dict(
            [[i, x] for (i, x) in zip(list(range(len(textlist))), textlist)]
        )

        self.active_key = default
        self.active_index = 0

        self.__fill()
        self.objcombo.clear()
        self.objcombo.set_model(self.store)
        cell = Gtk.CellRendererText()
        self.objcombo.pack_start(cell, True)
        self.objcombo.add_attribute(cell, "text", 1)
        self.objcombo.set_active(self.active_index)
        self.objcombo.connect("changed", self.on_combochange)

        # fill the entrybox with required data
        self.entry_reinit()
        self.objentry.connect("changed", self._on_change_entry)

        # set correct editable
        self.enable(not read_only)

    ##    def destroy(self):
    ##        """
    ##        Unset all elements that can prevent garbage collection
    ##        """
    ##        self.set_val_list = None
    ##        self.get_val_list = None
    ##        self.objcombo = None
    ##        self.objentry = None

    def __fill(self):
        """
        Fill combo with data
        """
        self.store = Gtk.ListStore(GObject.TYPE_INT, GObject.TYPE_STRING)
        keys = sorted(list(self.mapping.keys()), key=self.__by_value_key)

        for index, key in enumerate(keys):
            self.store.append(row=[key, self.mapping[key]])
            if key == self.active_key:
                self.active_index = index

    def __by_value(self, first, second):
        """
        Method for sorting keys based on the values.
        """
        fvalue = self.mapping[first]
        svalue = self.mapping[second]
        return glocale.strcoll(fvalue, svalue)

    def __by_value_key(self, first):
        """
        Method for sorting keys based on the values.
        """
        return glocale.sort_key(self.mapping[first])

    def on_combochange(self, obj):
        """
        callback for change on the combo, change active iter, update
        associated entrybox
        """
        self.active_key = self.store.get_value(self.objcombo.get_active_iter(), 0)
        self.entry_reinit()

    def reinit(self, set_val_list, get_val_list):
        """
        The interface is attached to another object, so the methods need to be
        reset.
        """
        self.set_val_list = set_val_list
        self.get_val_list = get_val_list
        self.update()

    def entry_reinit(self):
        """
        Make the entry field show the value corresponding to the active key
        """
        self.objentry.set_text(self.get_val_list[self.active_key]())
        self.set_val = self.set_val_list[self.active_key]
        self.get_val = self.get_val_list[self.active_key]

    def _on_change_entry(self, obj):
        """
        Callback when the entry field changes
        """
        self.set_val_list[self.active_key](self.get_value_entry())

    def get_value_entry(self):
        return str(self.objentry.get_text())

    def enable(self, value):
        self.objentry.set_sensitive(value)
        self.objentry.set_editable(value)

    def update(self):
        """
        Method called when object changed without interface change
        Eg: name editor save brings you back to person editor that must update
        """
        self.entry_reinit()


# -------------------------------------------------------------------------
#
# MonitoredTagList class
#
# -------------------------------------------------------------------------
class MonitoredTagList:
    """
    A MonitoredTagList consists of a label to display a list of tags and a
    button to invoke the tag editor.
    """

    def __init__(
        self, label, button, set_list, get_list, db, uistate, track, readonly=False
    ):
        self.uistate = uistate
        self.track = track
        self.db = db
        self.set_list = set_list

        self.tag_list = []
        for handle in get_list():
            tag = self.db.get_tag_from_handle(handle)
            if tag:
                self.tag_list.append((handle, tag.get_name()))

        self.all_tags = []
        for handle in self.db.get_tag_handles(sort_handles=True):
            tag = self.db.get_tag_from_handle(handle)
            self.all_tags.append((tag.get_handle(), tag.get_name()))

        self.label = label
        self.label.set_halign(Gtk.Align.START)
        self.label.set_ellipsize(Pango.EllipsizeMode.END)
        image = Gtk.Image()
        image.set_from_icon_name("gramps-tag", Gtk.IconSize.MENU)
        button.set_image(image)
        button.set_tooltip_text(_("Edit the tag list"))
        button.connect("button-press-event", self.cb_edit)
        button.connect("key-press-event", self.cb_edit)
        button.set_sensitive(not readonly)

        self._display()

    ##    def destroy(self):
    ##        """
    ##        Unset all elements that can prevent garbage collection
    ##        """
    ##        self.uistate = None
    ##        self.track = None
    ##        self.db = None
    ##        self.set_list = None

    def _display(self):
        """
        Display the tag list.
        """
        tag_text = ", ".join(item[1] for item in self.tag_list)
        self.label.set_text(tag_text)
        self.label.set_tooltip_text(tag_text)

    def cb_edit(self, button, event):
        """
        Invoke the tag editor.
        """
        if event.type == Gdk.EventType.BUTTON_PRESS or (
            event.type == Gdk.EventType.KEY_PRESS
            and event.keyval in (_RETURN, _KP_ENTER)
        ):
            from ..editors import EditTagList

            editor = EditTagList(self.tag_list, self.all_tags, self.uistate, self.track)
            if editor.return_list is not None:
                self.tag_list = editor.return_list
                self._display()
                self.set_list([item[0] for item in self.tag_list])
            return True
        return False



# -*- coding: utf-8 -*-
#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2005-2007  Donald N. Allingham
# Copyright (C) 2008-2009  Gary Burton
# Copyright (C) 2009-2012  Doug Blank <doug.blank@gmail.com>
# Copyright (C) 2011       Tim G L Lyons
# Copyright (C) 2015-      Serge Noiraud
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

"""
This package implements access to Gramps configuration.
"""

# ---------------------------------------------------------------
#
# Gramps imports
#
# ---------------------------------------------------------------
import os
import re
import logging

# ---------------------------------------------------------------
#
# Gramps imports
#
# ---------------------------------------------------------------
from .const import USER_CONFIG, USER_DATA, USER_HOME, VERSION_DIR
from .utils.configmanager import ConfigManager
from .const import GRAMPS_LOCALE as glocale

_ = glocale.translation.gettext


# _T_ is a gramps-defined keyword -- see po/update_po.py and po/genpot.sh
def _T_(value, context=""):  # enable deferred translations
    return "%s\x04%s" % (context, value) if context else value


# ---------------------------------------------------------------
#
# Constants
#
# ---------------------------------------------------------------
INIFILE = os.path.join(VERSION_DIR, "gramps.ini")


# ---------------------------------------------------------------
#
# Module functions
#
# ---------------------------------------------------------------
def register(key, value):
    """Module shortcut to register key, value"""
    return CONFIGMAN.register(key, value)


def get(key):
    """Module shortcut to get value from key"""
    return CONFIGMAN.get(key)


def get_default(key):
    """Module shortcut to get default from key"""
    return CONFIGMAN.get_default(key)


def has_default(key):
    """Module shortcut to get see if there is a default for key"""
    return CONFIGMAN.has_default(key)


def get_sections():
    """Module shortcut to get all section names of settings"""
    return CONFIGMAN.get_sections()


def get_section_settings(section):
    """Module shortcut to get all settings of a section"""
    return CONFIGMAN.get_section_settings(section)


def set(key, value):
    """Module shortcut to set value from key"""
    return CONFIGMAN.set(key, value)


def is_set(key):
    """Module shortcut to set value from key"""
    return CONFIGMAN.is_set(key)


def save(filename=None):
    """Module shortcut to save config file"""
    return CONFIGMAN.save(filename)


def connect(key, func):
    """
    Module shortcut to connect a key to a callback func.
    Returns a unique callback ID number.
    """
    return CONFIGMAN.connect(key, func)


def disconnect(callback_id):
    """Module shortcut to remove callback by ID number"""
    return CONFIGMAN.disconnect(callback_id)


def reset(key=None):
    """Module shortcut to reset some or all config data"""
    return CONFIGMAN.reset(key)


def load(filename=None, oldstyle=False):
    """Module shortcut to load an INI file into config data"""
    return CONFIGMAN.load(filename, oldstyle)


def emit(key):
    """Module shortcut to call all callbacks associated with key"""
    return CONFIGMAN.emit(key)


# ---------------------------------------------------------------
#
# Register the system-wide settings in a singleton config manager
#
# ---------------------------------------------------------------

CONFIGMAN = ConfigManager(INIFILE, "plugins")

register("behavior.addmedia-image-dir", "")
register("behavior.addmedia-relative-path", False)
register("behavior.autoload", False)
register("behavior.avg-generation-gap", 20)
register("behavior.check-for-addon-updates", 0)
register("behavior.check-for-addon-update-types", ["new"])
register("behavior.last-check-for-addon-updates", "1970/01/01")
register("behavior.previously-seen-addon-updates", [])
register("behavior.do-not-show-previously-seen-addon-updates", True)
register("behavior.date-about-range", 50)
register("behavior.date-after-range", 50)
register("behavior.date-before-range", 50)
register("behavior.generation-depth", 15)
register("behavior.max-age-prob-alive", 110)
register("behavior.max-sib-age-diff", 20)
register("behavior.min-generation-years", 13)
register("behavior.owner-warn", False)
register("behavior.immediate-warn", False)
register("behavior.pop-plugin-status", False)
register("behavior.recent-export-type", 3)
register("behavior.runcheck", False)
register("behavior.spellcheck", False)
register("behavior.startup", 0)
register("behavior.surname-guessing", 0)
register("behavior.translator-needed", True)
register("behavior.use-tips", False)
register("behavior.welcome", 100)
register("behavior.web-search-url", "http://google.com/#&q=%(text)s")
register(
    "behavior.addons-url",
    "https://raw.githubusercontent.com/gramps-project/addons/master/gramps52",
)
register(
    "behavior.addons-projects",
    [
        [
            "Gramps",
            "https://raw.githubusercontent.com/gramps-project/addons/master/gramps52",
            True,
        ]
    ],
)
register("behavior.addons-allow-install", False)

register("csv.dialect", "excel")
register("csv.delimiter", ",")

register("database.backend", "sqlite")
register("database.compress-backup", True)
register("database.backup-path", USER_HOME)
register("database.backup-on-exit", True)
register("database.autobackup", 0)
register("database.path", os.path.join(USER_DATA, "grampsdb"))
register("database.host", "")
register("database.port", "")

register(
    "export.proxy-order",
    [["privacy", 0], ["living", 0], ["person", 0], ["note", 0], ["reference", 0]],
)

register("geography.center-lon", 0.0)
register("geography.lock", False)
register("geography.center-lat", 0.0)
register("geography.map_service", 1)
register("geography.zoom", 0)
register("geography.zoom_when_center", 12)
register("geography.show_cross", False)
register("geography.path", "")
register("geography.use-keypad", True)
register("geography.personal-map", "")

# note that other calls to "register" are done in realtime (when
# needed), for instance to four 'interface.clipboard' variables --
# so do a recursive grep for "setup_configs" to see all the (base) names
register("interface.dont-ask", False)
register(
    "interface.view-categories",
    [
        "Dashboard",
        "People",
        "Relationships",
        "Families",
        "Ancestry",
        "Events",
        "Places",
        "Geography",
        "Sources",
        "Citations",
        "Repositories",
        "Media",
        "Notes",
    ],
)
register("interface.filter", False)
register("interface.fullscreen", False)
register("interface.grampletbar-close", False)
register("interface.ignore-gexiv2", False)
register("interface.ignore-pil", False)
register("interface.ignore-osmgpsmap", False)
register("interface.main-window-height", 500)
register("interface.main-window-horiz-position", 15)
register("interface.main-window-vert-position", 10)
register("interface.main-window-width", 775)
register("interface.mapservice", "OpenStreetMap")
register("interface.open-with-default-viewer", False)
register("interface.pedview-layout", 0)
register("interface.pedview-show-images", True)
register("interface.pedview-show-marriage", False)
register("interface.pedview-tree-size", 5)
register("interface.pedview-tree-direction", 2)
register("interface.pedview-show-unknown-people", False)
register("interface.place-name-height", 100)
register("interface.place-name-width", 450)
register("interface.sidebar-text", True)
register("interface.size-checked", False)
register("interface.statusbar", 1)
register("interface.toolbar-on", True)
register("interface.toolbar-text", False)
register("interface.hide-lds", False)
register("interface.toolbar-clipboard", True)
register("interface.toolbar-addons", True)
register("interface.toolbar-preference", True)
register("interface.toolbar-reports", True)
register("interface.toolbar-tools", True)
register("interface.view", True)
register("interface.surname-box-height", 150)
register("interface.treemodel-cache-size", 1000)

register("paths.recent-export-dir", USER_HOME)
register("paths.recent-file", "")
register("paths.recent-import-dir", USER_HOME)
register("paths.report-directory", USER_HOME)
register("paths.website-directory", USER_HOME)
register("paths.website-cms-uri", "")
register("paths.website-cal-uri", "")
register("paths.website-extra-page-uri", "")
register("paths.website-extra-page-name", "")
register("paths.quick-backup-directory", USER_HOME)
register(
    "paths.quick-backup-filename",
    "%(filename)s_%(year)d-%(month)02d-%(day)02d.%(extension)s",
)

register("preferences.quick-backup-include-mode", False)
register("preferences.date-format", 0)
register("preferences.calendar-format-report", 0)
register("preferences.calendar-format-input", 0)
register("preferences.february-29", 0)  # 0: 02/28; 1: 03/01; 2: only the 02/29
register("preferences.cprefix", "C%04d")
register("preferences.default-source", False)
register("preferences.tag-on-import", False)
register("preferences.tag-on-import-format", _("Imported %Y/%m/%d %H:%M:%S"))
register("preferences.eprefix", "E%04d")
register("preferences.family-warn", True)
register("preferences.fprefix", "F%04d")
register("preferences.hide-ep-msg", False)
register("preferences.invalid-date-format", "<b>%s</b>")
register("preferences.iprefix", "I%04d")
register("preferences.name-format", 1)
register("preferences.place-format", 0)
register("preferences.place-auto", True)
register("preferences.coord-format", 0)
register("preferences.patronimic-surname", False)
register("preferences.no-given-text", "[%s]" % _("Missing Given Name"))
register("preferences.no-record-text", "[%s]" % _("Missing Record"))
register("preferences.no-surname-text", "[%s]" % _("Missing Surname"))
register("preferences.nprefix", "N%04d")
register("preferences.online-maps", False)
register("preferences.oprefix", "O%04d")
register("preferences.paper-metric", 0)
register("preferences.paper-preference", "Letter")
register("preferences.pprefix", "P%04d")
register("preferences.private-given-text", "%s" % _T_("[Living]"))
register("preferences.private-record-text", "[%s]" % _("Private Record"))
register("preferences.private-surname-text", "%s" % _T_("[Living]"))
register("preferences.rprefix", "R%04d")
register("preferences.sprefix", "S%04d")
register("preferences.use-last-view", False)
register("preferences.last-view", "")
register("preferences.last-views", [])
register("preferences.family-relation-type", 3)  # UNKNOWN
register("preferences.age-display-precision", 1)
register("preferences.age-after-death", True)
register("preferences.age-rounded-year", True)
register("preferences.cite-plugin", "cite-legacy")

register("colors.scheme", 0)
register("colors.male-alive", ["#b8cee6", "#1f344a"])
register("colors.male-dead", ["#b8cee6", "#2d3039"])
register("colors.female-alive", ["#feccf0", "#62242D"])
register("colors.female-dead", ["#feccf0", "#3a292b"])
register("colors.other-alive", ["#94ef9e", "#285b27"])
register("colors.other-dead", ["#94ef9e", "#062304"])
register("colors.unknown-alive", ["#f3dbb6", "#75507B"])
register("colors.unknown-dead", ["#f3dbb6", "#35103b"])
register("colors.family", ["#eeeeee", "#454545"])
register("colors.family-married", ["#eeeeee", "#454545"])
register("colors.family-unmarried", ["#eeeeee", "#454545"])
register("colors.family-civil-union", ["#eeeeee", "#454545"])
register("colors.family-unknown", ["#eeeeee", "#454545"])
register("colors.family-divorced", ["#ffdede", "#5c3636"])
register("colors.home-person", ["#bbe68a", "#304918"])
register("colors.border-male-alive", ["#1f4986", "#171d26"])
register("colors.border-male-dead", ["#000000", "#000000"])
register("colors.border-female-alive", ["#861f69", "#261111"])
register("colors.border-female-dead", ["#000000", "#000000"])
register("colors.border-other-alive", ["#2a5f16", "#26a269"])
register("colors.border-other-dead", ["#000000", "#000000"])
register("colors.border-unknown-alive", ["#8e5801", "#8e5801"])
register("colors.border-unknown-dead", ["#000000", "#000000"])
register("colors.border-family", ["#cccccc", "#252525"])
register("colors.border-family-divorced", ["#ff7373", "#720b0b"])

register("researcher.researcher-addr", "")
register("researcher.researcher-locality", "")
register("researcher.researcher-city", "")
register("researcher.researcher-country", "")
register("researcher.researcher-email", "")
register("researcher.researcher-name", "")
register("researcher.researcher-phone", "")
register("researcher.researcher-postal", "")
register("researcher.researcher-state", "")

register("plugin.hiddenplugins", [])
register("plugin.addonplugins", [])

register("utf8.in-use", False)
register("utf8.selected-font", "")
register("utf8.death-symbol", 2)
register("utf8.birth-symbol", "*")
register("utf8.baptism-symbol", "~")
register("utf8.marriage-symbol", "oo")
register("utf8.engaged-symbol", "o")
register("utf8.divorce-symbol", "o|o")
register("utf8.partner-symbol", "o-o")
register("utf8.dead-symbol", "✝")
register("utf8.buried-symbol", "[]")
register("utf8.cremated-symbol", "⚱")
register("utf8.killed-symbol", "x")

if __debug__:  # enable a simple CLI test to see if the datestrings exist
    register("test.january", _("January", "localized lexeme inflections"))

# ---------------------------------------------------------------
#
# Upgrade Conversions go here.
#
# ---------------------------------------------------------------

# If we have not already upgraded to this version,
# we can tell by seeing if there is a key file for this version:
if not os.path.exists(CONFIGMAN.filename):
    # If not, let's read old if there:
    if os.path.exists(os.path.join(USER_CONFIG, "keys.ini")):
        # read it in old style:
        logging.info("Importing old key file 'keys.ini'...")
        CONFIGMAN.load(os.path.join(USER_CONFIG, "keys.ini"), oldstyle=True)
        logging.info("Done importing old key file 'keys.ini'")
    # other version upgrades here...
    # check previous version of gramps:
    fullpath, filename = os.path.split(CONFIGMAN.filename)
    fullpath, previous = os.path.split(fullpath)
    match = re.match(r"gramps(\d*)", previous)
    if match:
        # cycle back looking for previous versions of gramps
        for i in range(1, 20):  # check back 2 gramps versions
            # -----------------------------------------
            # TODO: Assumes minor version is a decimal, not semantic versioning
            #       Uses ordering ... 4.9, 5.0, 5.1, ...
            #       Not ... 4.9, 4.10, 4.11, 5.0, 5.1, ...
            # If not true, need to add a different method to auto upgrade.
            # Perhaps addings specific list of versions to check
            # -----------------------------------------
            digits = str(int(match.groups()[0]) - i)
            previous_grampsini = os.path.join(fullpath, "gramps" + digits, filename)
            if os.path.exists(previous_grampsini):
                logging.info("Importing old config file '%s'...", previous_grampsini)
                CONFIGMAN.load(previous_grampsini)
                logging.info("Done importing old config file '%s'", previous_grampsini)
                break

# ---------------------------------------------------------------
#
# Now, load the settings from the config file, if one
#
# ---------------------------------------------------------------
CONFIGMAN.load()

config = CONFIGMAN
if config.get("database.backend") == "bsddb":
    config.set("database.backend", "sqlite")



# -*- coding: utf-8 -*-
#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2023       John Ralls
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
"""
win32locale provides l18n setup for the Microsoft Windows platform.
"""
import os
import locale
from ctypes import cdll
import logging

LOG = logging.getLogger("." + __name__)
LOG.propagate = True


# Map of languages for converting to Microsoft locales and naming
# locales for display to the user.  It's important to add to this list
# when a new translation is added.  Note the dummy _(): That's just to
# get xgettext to include the string in gramps.pot; actual translation
# is done in _get_language_string() below.
# (The gramps officially-supported language list is ALL_LINGUAS in setup.py)
def _(msgid):
    return msgid


_LOCALE_NAMES = {
    "ar": ("Arabic_Saudi Arabia", "1256", _("Arabic")),
    "bg": ("Bulgrian_Bulgaria", "1251", _("Bulgarian")),
    # Windows has no translation for Breton
    "br": (None, None, _("Breton")),
    "ca": ("Catalan_Spain", "1252", _("Catalan")),
    "cs": ("Czech_Czech Republic", "1250", _("Czech")),
    "da": ("Danish_Denmark", "1252", _("Danish")),
    "de": ("German_Germany", "1252", _("German")),
    "de_AT": ("German_Austria", "1252", _("German (Austria)")),
    "el": ("Greek_Greece", "1253", _("Greek")),
    "en": ("English_United States", "1252", _("English (USA)")),
    "en_GB": ("English_United Kingdom", "1252", _("English")),
    # Windows has no translation for Esperanto
    "eo": (None, None, _("Esperanto")),
    "es": ("Spanish_Spain", "1252", _("Spanish")),
    "fa": ("Farsi_Iran", "1256", _("Farsi")),
    "fi": ("Finnish_Finland", "1252", _("Finnish")),
    "fr": ("French_France", "1252", _("French")),
    # Windows has no translation for Gaelic
    "ga": (None, None, _("Gaelic")),
    "gl": ("Galician_Spain", "1252", _("Galician")),
    "he": ("Hebrew_Israel", "1255", _("Hebrew")),
    "hr": ("Croatian_Croatia", "1250", _("Croatian")),
    "hu": ("Hungarian_Hungary", "1250", _("Hungarian")),
    "id": ("Indonesian", "1057", _("Indonesian")),
    "is": ("Icelandic", "1252", _("Icelandic")),
    "it": ("Italian_Italy", "1252", _("Italian")),
    "ja": ("Japanese_Japan", "932", _("Japanese")),
    "ko": ("Korean_Korea", "949", _("Korean")),
    "lt": ("Lithuanian_Lithuania", "1252", _("Lithuanian")),
    # Windows has no translation for Macedonian
    "mk": (None, None, _("Macedonian")),
    "nb": ("Norwegian_Norway", "1252", _("Norwegian Bokmal")),
    "nl": ("Dutch_Netherlands", "1252", _("Dutch")),
    "nn": ("Norwegian-Nynorsk_Norway", "1252", _("Norwegian Nynorsk")),
    "pl": ("Polish_Poland", "1250", _("Polish")),
    "pt_BR": ("Portuguese_Brazil", "1252", _("Portuguese (Brazil)")),
    "pt_PT": ("Portuguese_Portugal", "1252", _("Portuguese (Portugal)")),
    "ro": ("Romanian_Romania", "1250", _("Romanian")),
    "ru": ("Russian_Russia", "1251", _("Russian")),
    "sk": ("Slovak_Slovakia", "1250", _("Slovak")),
    "sl": ("Slovenian_Slovenia", "1250", _("Slovenian")),
    "sq": ("Albanian_Albania", "1250", _("Albanian")),
    "sr": ("Serbian(Cyrillic)_Serbia and Montenegro", "1251", _("Serbian")),
    "sv": ("Swedish_Sweden", "1252", _("Swedish")),
    # Windows has no codepage for Tamil
    "ta": (None, None, _("Tamil")),
    "tr": ("Turkish_Turkey", "1254", _("Turkish")),
    "uk": ("Ukrainian_Ukraine", "1251", _("Ukrainian")),
    "vi": ("Vietnamese_Vietnam", "1258", _("Vietnamese")),
    "zh_CN": ("Chinese_China", "936", _("Chinese (Simplified)")),
    "zh_HK": ("Chinese_Hong Kong", "950", _("Chinese (Hong Kong)")),
    "zh_TW": ("Chinese_Taiwan", "950", _("Chinese (Traditional)")),
}


def _check_mswin_locale(loc):
    msloc = None
    try:
        msloc = _LOCALE_NAMES[loc[:5]][:2]
        newloc = loc[:5]
    except KeyError:
        try:
            msloc = _LOCALE_NAMES[loc[:2]][:2]
            newloc = loc[:2]
        except KeyError:
            # US English is the outlier, all other English locales want
            # real English:
            if loc[:2] == ("en") and loc[:5] != "en_US":
                return ("en_GB", "1252")
            return (None, None)
    return (newloc, msloc)


def _check_mswin_locale_reverse(loc):
    (lang, country) = loc.split("_")
    # US English is the outlier, all other English locales want real English:
    if lang == "English" and country != "United States":
        return ("en_GB", "1252")

    for newloc, msloc in _LOCALE_NAMES.items():
        if not msloc[0] is None and loc == msloc[0]:
            return (newloc, msloc[1])

    # Didn't get a full language-country match, find the first locale
    # with the same language:
    for newloc, msloc in _LOCALE_NAMES.items():
        if loc == msloc[2] or lang == msloc[2]:
            return (newloc, msloc[1])

    return (None, None)


def _set_lang(glocale):
    if "LANG" in os.environ:
        (lang, loc) = _check_mswin_locale(os.environ["LANG"])
        if loc:
            locale.setlocale(locale.LC_ALL, ".".join(loc))
            glocale.lang = lang
            glocale.encoding = loc[1]
        else:
            LOG.debug("%%LANG%% value %s not usable", os.environ["LANG"])
    if not glocale.lang:
        locale.setlocale(locale.LC_ALL, "")
        (pylang, pyenc) = locale.getlocale()
        # Sometimes getlocale returns a Microsoft locale code
        (lang, enc) = _check_mswin_locale_reverse(pylang)
        if not lang is None:
            glocale.lang = lang
            glocale.encoding = enc
        else:  # and sometimes it returns a POSIX locale
            (lang, enc) = _check_mswin_locale(pylang)
            if not lang is None:
                glocale.lang = lang
                glocale.encoding = enc
            else:
                LOG.debug("No usable locale found in environment")

    if not glocale.lang:
        glocale.lang = "C"
        glocale.encoding = "cp1252"


def _set_languages(glocale):
    lang = None
    if "LC_MESSAGES" in os.environ:
        lang = glocale.check_available_translations(os.environ["LC_MESSAGES"])
    if lang:
        glocale.language = [lang]
    else:
        LOG.debug("No translation for %%LC_MESSAGES%% locale")
    if "LANGUAGE" in os.environ:
        language = [
            x
            for x in [
                glocale.check_available_translations(l)
                for l in os.environ["LANGUAGE"].split(":")
            ]
            if x
        ]
        if language:
            glocale.language = language
        else:
            LOG.debug("No languages with translations found " "in %%LANGUAGES%%")
    if not glocale.language:
        glocale.language = [glocale.lang[:5]]


def _set_collation(glocale, have_icu, icu_locales):
    coll = None
    if "COLLATION" in os.environ:
        coll = os.environ["COLLATION"]
        if have_icu:
            if coll[:2] in icu_locales:
                glocale.collation = coll
            else:
                glocale.collation = glocale.lang
        else:
            (coll, loc) = _check_mswin_locale(coll)
            if not loc:
                (coll, loc) = _check_mswin_locale(glocale.lang)
                glocale.collation = ".".join(loc)
                locale.setlocale(locale.LC_COLLATE, glocale.collation)
    else:
        if have_icu:
            glocale.collation = glocale.lang
        else:
            (coll, loc) = _check_mswin_locale(glocale.lang)
            if loc:
                glocale.collation = ".".join(loc)
            else:
                glocale.collation = "C"
                locale.setlocale(locale.LC_COLLATE, glocale.collation)


def _set_calendar(glocale):
    # We can't import datahandler stuff or we'll get a circular
    # dependency, so we rely on the available translations list
    if "LC_TIME" in os.environ:
        glocale.calendar = (
            glocale.check_available_translations(os.environ["LC_TIME"]) or glocale.lang
        )
    else:
        glocale.calendar = glocale.lang


def win32_locale_init_from_env(glocale, have_icu, icu_locales):
    """
    The Windows implementation of Python ignores environment
    variables when setting the locale; it only pays attention to
    the control panel language settings -- which for practical
    purposes limits one to the language for which one purchased
    Windows. This function enables using alternative
    localizations.
    """
    _set_lang(glocale)
    _set_languages(glocale)
    _set_collation(glocale, have_icu, icu_locales)
    _set_calendar(glocale)


def win32_locale_bindtextdomain(localedomain, localedir):
    """
    Help routine for loading and setting up libintl attributes
    Returns libintl
    """
    try:
        libintl = cdll.LoadLibrary("libintl-8")
        libintl.bindtextdomain(localedomain, localedir)
        libintl.textdomain(localedomain)
        libintl.bind_textdomain_codeset(localedomain, "UTF-8")

    except WindowsError:
        LOG.warning(
            "Localization library libintl not on %PATH%, "
            "localization will be incomplete"
        )



# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2009-2010  Nick Hall
# Copyright (C) 2011       Tim G L Lyons
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

"""
Citation Tree View (or Source tree view).
A view showing all the Sources with child Citations
"""
# -------------------------------------------------------------------------
#
# python modules
#
# -------------------------------------------------------------------------
import logging

LOG = logging.getLogger(".citation")
_LOG = logging.getLogger(".gui.citationtreeview")

# -------------------------------------------------------------------------
#
# GTK/Gnome modules
#
# -------------------------------------------------------------------------
from gi.repository import Gtk

# -------------------------------------------------------------------------
#
# Gramps modules
#
# -------------------------------------------------------------------------
from gramps.gui.views.listview import TEXT, MARKUP, ICON
from gramps.gui.views.treemodels.citationtreemodel import CitationTreeModel
from gramps.gen.plug import CATEGORY_QR_SOURCE_OR_CITATION
from gramps.gen.lib import Citation, Source
from gramps.gui.views.listview import ListView
from gramps.gui.views.bookmarks import CitationBookmarks
from gramps.gen.errors import WindowActiveError, HandleError
from gramps.gui.ddtargets import DdTargets
from gramps.gui.dialog import ErrorDialog
from gramps.gui.editors import EditCitation, EditSource
from gramps.gui.filters.sidebar import SourceSidebarFilter
from gramps.gui.merge import MergeCitation, MergeSource
from gramps.plugins.lib.libsourceview import LibSourceView

# -------------------------------------------------------------------------
#
# Internationalization
#
# -------------------------------------------------------------------------
from gramps.gen.const import GRAMPS_LOCALE as glocale

_ = glocale.translation.sgettext


# -------------------------------------------------------------------------
#
# PlaceTreeView
#
# -------------------------------------------------------------------------
class CitationTreeView(LibSourceView, ListView):
    """
    A hierarchical view of sources with citations below them.
    """

    # The data items here have to correspond, in order, to the items in
    # src/giu/views/treemodels/citationtreemodel.py
    COL_TITLE_PAGE = 0
    COL_ID = 1
    COL_DATE = 2
    COL_CONFIDENCE = 3
    COL_PRIV = 4
    COL_TAGS = 5
    COL_CHAN = 6
    COL_SRC_AUTH = 7
    COL_SRC_ABBR = 8
    COL_SRC_PINFO = 9
    # column definitions
    COLUMNS = [
        (_("Source: Title or Citation: Volume/Page"), TEXT, None),
        (_("ID"), TEXT, None),
        (_("Date"), MARKUP, None),
        (_("Confidence"), TEXT, None),
        (_("Private"), ICON, "gramps-lock"),
        (_("Tags"), TEXT, None),
        (_("Last Changed"), TEXT, None),
        (_("Source: Author"), TEXT, None),
        (_("Source: Abbreviation"), TEXT, None),
        (_("Source: Publication Information"), TEXT, None),
    ]
    COLUMN_FILTERABLE = [
        COL_TITLE_PAGE,
        COL_ID,
        COL_CHAN,
        COL_SRC_AUTH,
        COL_SRC_ABBR,
        COL_SRC_PINFO,
    ]
    # default setting with visible columns, order of the col, and their size
    CONFIGSETTINGS = (
        ("columns.visible", [COL_TITLE_PAGE, COL_ID, COL_SRC_AUTH, COL_SRC_PINFO]),
        (
            "columns.rank",
            [
                COL_TITLE_PAGE,
                COL_ID,
                COL_DATE,
                COL_CONFIDENCE,
                COL_PRIV,
                COL_TAGS,
                COL_CHAN,
                COL_SRC_AUTH,
                COL_SRC_ABBR,
                COL_SRC_PINFO,
            ],
        ),
        ("columns.size", [200, 75, 100, 75, 40, 100, 100, 150, 100, 150]),
    )
    ADD_MSG = _("Add a new citation and a new source")
    ADD_SOURCE_MSG = _("Add a new source")
    ADD_CITATION_MSG = _("Add a new citation to an existing source")
    EDIT_MSG = _("Edit the selected citation or source")
    DEL_MSG = _("Delete the selected citation or source")
    MERGE_MSG = _("Merge the selected citations or selected sources")
    FILTER_TYPE = "Citation"
    QR_CATEGORY = CATEGORY_QR_SOURCE_OR_CITATION

    def __init__(self, pdata, dbstate, uistate, nav_group=0):
        signal_map = {
            "citation-add": self._citation_row_add,
            "citation-update": self._citation_row_update,
            "citation-delete": self._citation_row_delete,
            "citation-rebuild": self._citation_object_build,
            "source-add": self._source_row_add,
            "source-update": self._source_row_update,
            "source-delete": self._source_row_delete,
            "source-rebuild": self._source_object_build,
        }

        ListView.__init__(
            self,
            _("Citation Tree View"),
            pdata,
            dbstate,
            uistate,
            CitationTreeModel,
            signal_map,
            CitationBookmarks,
            nav_group,
            multiple=True,
            filter_class=SourceSidebarFilter,
        )

        self.additional_uis.append(self.additional_ui)

    # Override change_active from NavigationView, so that only Citations can be
    # put in the history list for the CitationTreeView
    def change_active(self, handle):
        try:
            self.dbstate.db.get_citation_from_handle(handle)
            super().change_active(handle)
        except HandleError:
            # FIXME: See http://www.gramps-project.org/bugs/view.php?id=6352 a
            # more comprehensive solution is needed in the long term. See also
            # add in Bookmarks.CitationBookmarks
            pass

    def setup_filter(self):
        """
        Override the setup of the default Search Bar in listview, so that only
        the searchable source fields are shown.
        """

        self.search_bar.setup_filter(
            [
                (self.COLUMNS[pair[1]][0], pair[1], pair[1] in self.exact_search())
                for pair in self.column_order()
                if pair[0] and pair[1] in self.COLUMN_FILTERABLE
            ]
        )

    def _print_handles(self, text, handle_list):
        for handle in handle_list:
            source, citation = self.get_source_or_citation(handle, False)
            _LOG.debug("\n\n\n")
            if source:
                _LOG.debug("---- %s -- source %s" % (text, source.get_title()))
            elif citation:
                _LOG.debug("---- %s -- citation %s" % (text, citation.get_page()))
            else:
                _LOG.debug("---- %s -- handle %s" % (text, handle))

    def _citation_row_add(self, handle_list):
        self._print_handles("citation row add", handle_list)
        self.row_add(handle_list)

    def _citation_row_update(self, handle_list):
        self._print_handles("citation row update", handle_list)
        self.row_update(handle_list)

    def _citation_row_delete(self, handle_list):
        self._print_handles("citation row delete", handle_list)
        self.row_delete(handle_list)

    def _citation_object_build(self, *args):
        _LOG.debug("citation object build")
        self.object_build(*args)

    def _source_row_add(self, handle_list):
        self._print_handles("source row add", handle_list)
        self.row_add(handle_list)

    def _source_row_update(self, handle_list):
        self._print_handles("source row update", handle_list)
        # if the source update changes the title or other item being sorted
        # then it may change position on tree; it's easier to just rebuild the
        # whole tree.  row_update cannot fix changes to first level of tree
        # self.row_update(handle_list)
        self.dirty = True
        self.build_tree()

    def _source_row_delete(self, handle_list):
        self._print_handles("source row delete", handle_list)
        self.row_delete(handle_list)

    def _source_object_build(self, *args):
        _LOG.debug("source object build")
        self.object_build(*args)

    def navigation_type(self):
        return "Citation"

    def object_build(self, *args):
        """
        Called when the tree must be rebuilt and bookmarks redrawn.
        """
        self.dirty = True
        if self.active:
            # Save the currently selected handles, if any:
            selected_ids = self.selected_handles()
            self.bookmarks.redraw()
            self.build_tree()
            # Reselect one, if it still exists after rebuild:
            for handle in selected_ids:
                # Still exist?  It might be either a source or citation handle.
                if self.dbstate.db.has_citation_handle(
                    handle
                ) or self.dbstate.db.has_source_handle(handle):
                    # Select it, and stop selecting:
                    self.change_active(handle)
                    break

    def drag_info(self):
        # Since drag only needs to work when just one row is selected, ideally,
        # this should just return SOURCE_LINK if one source is selected and
        # CITATION_LINK if one citation is selected, and probably None
        # otherwise. However, this doesn't work. Drag and drop failed to work at
        # all for citationtree view, and I think this was because None is
        # returned during initialisation. There is also a problem where it seems
        # at some point during a citation merge, neither a Source nor a Citation
        # is selected. Hence the simplistic solution implemented below, where
        # CITATION_LINK is always returned except when it is obviously correct
        # to return SOURCE_LINK.

        selection = self.selected_handles()
        if len(selection) == 1:
            try:
                self.dbstate.db.get_source_from_handle(selection[0])
                return DdTargets.SOURCE_LINK
            except HandleError:
                pass
        return DdTargets.CITATION_LINK

    def get_stock(self):
        return "gramps-citation"

    def get_viewtype_stock(self):
        """
        Override the default icon.  Set for hierarchical view.
        """
        return "gramps-tree-group"

    def define_actions(self):
        """
        This defines the possible actions for the citation views.
        Possible actions are:
        add_source: Add a new source (this is also available from the
                      source view)
        add:        Add a new citation and a new source (this can also be done
                      by source view add a source, then citation view add a new
                      citation to an existing source)
        share:      Add a new citation to an existing source (when a source is
                      selected)
        edit:       Edit a source or a citation.
        merge:      Merge the selected sources or citations.
        remove:     Delete the selected sources or citations.


        """
        ListView.define_actions(self)

        self.action_list.extend(
            [
                ("AddSource", self.add_source),
                ("AddCitation", self.share),
                ("OpenAllNodes", self.open_all_nodes),
                ("CloseAllNodes", self.close_all_nodes),
            ]
        )

    additional_ui = [  # Defines the UI string for UIManager
        """
      <placeholder id="LocalExport">
        <item>
          <attribute name="action">win.ExportTab</attribute>
          <attribute name="label" translatable="yes">Export View...</attribute>
        </item>
      </placeholder>
""",
        """
      <section id="AddEditBook">
        <item>
          <attribute name="action">win.AddBook</attribute>
          <attribute name="label" translatable="yes">_Add Bookmark</attribute>
        </item>
        <item>
          <attribute name="action">win.EditBook</attribute>
          <attribute name="label" translatable="no">%s...</attribute>
        </item>
      </section>
"""
        % _("Organize Bookmarks"),
        """
      <placeholder id="CommonGo">
      <section>
        <item>
          <attribute name="action">win.Back</attribute>
          <attribute name="label" translatable="yes">_Back</attribute>
        </item>
        <item>
          <attribute name="action">win.Forward</attribute>
          <attribute name="label" translatable="yes">_Forward</attribute>
        </item>
      </section>
      </placeholder>
""",
        """
      <section id='CommonEdit' groups='RW'>
        <item>
          <attribute name="action">win.Add</attribute>
          <attribute name="label" translatable="yes">_Add...</attribute>
        </item>
        <item>
          <attribute name="action">win.AddSource</attribute>
          <attribute name="label" translatable="yes">Add source...</attribute>
        </item>
        <item>
          <attribute name="action">win.AddCitation</attribute>
          <attribute name="label" translatable="yes">Add citation...</attribute>
        </item>
        <item>
          <attribute name="action">win.Edit</attribute>
          <attribute name="label">%s</attribute>
        </item>
        <item>
          <attribute name="action">win.Remove</attribute>
          <attribute name="label" translatable="yes">_Delete</attribute>
        </item>
        <item>
          <attribute name="action">win.Merge</attribute>
          <attribute name="label" translatable="yes">_Merge...</attribute>
        </item>
      </section>
"""
        % _("_Edit...", "action"),  # to use sgettext()
        """
        <placeholder id='otheredit'>
        <item>
          <attribute name="action">win.FilterEdit</attribute>
          <attribute name="label" translatable="yes">"""
        """Citation Filter Editor</attribute>
        </item>
        </placeholder>
""",  # Following are the Toolbar items
        """
    <placeholder id='CommonNavigation'>
    <child groups='RO'>
      <object class="GtkToolButton">
        <property name="icon-name">go-previous</property>
        <property name="action-name">win.Back</property>
        <property name="tooltip_text" translatable="yes">"""
        """Go to the previous object in the history</property>
        <property name="label" translatable="yes">_Back</property>
        <property name="use-underline">True</property>
      </object>
      <packing>
        <property name="homogeneous">False</property>
      </packing>
    </child>
    <child groups='RO'>
      <object class="GtkToolButton">
        <property name="icon-name">go-next</property>
        <property name="action-name">win.Forward</property>
        <property name="tooltip_text" translatable="yes">"""
        """Go to the next object in the history</property>
        <property name="label" translatable="yes">_Forward</property>
        <property name="use-underline">True</property>
      </object>
      <packing>
        <property name="homogeneous">False</property>
      </packing>
    </child>
    </placeholder>
""",
        """
    <placeholder id='BarCommonEdit'>
    <child groups='RW'>
      <object class="GtkToolButton">
        <property name="icon-name">list-add</property>
        <property name="action-name">win.Add</property>
        <property name="tooltip_text">%s</property>
        <property name="label" translatable="yes">_Add...</property>
        <property name="use-underline">True</property>
      </object>
      <packing>
        <property name="homogeneous">False</property>
      </packing>
    </child>
    <child groups='RW'>
      <object class="GtkToolButton">
        <property name="icon-name">gramps-source</property>
        <property name="action-name">win.AddSource</property>
        <property name="tooltip_text">%s</property>
        <property name="label" translatable="yes">Add source...</property>
      </object>
      <packing>
        <property name="homogeneous">False</property>
      </packing>
    </child>
    <child groups='RW'>
      <object class="GtkToolButton">
        <property name="icon-name">gramps-citation</property>
        <property name="action-name">win.AddCitation</property>
        <property name="tooltip_text">%s</property>
        <property name="label" translatable="yes">Add citation...</property>
      </object>
      <packing>
        <property name="homogeneous">False</property>
      </packing>
    </child>
    <child groups='RW'>
      <object class="GtkToolButton">
        <property name="icon-name">gtk-edit</property>
        <property name="action-name">win.Edit</property>
        <property name="tooltip_text">%s</property>
        <property name="label" translatable="yes">Edit...</property>
        <property name="use-underline">True</property>
      </object>
      <packing>
        <property name="homogeneous">False</property>
      </packing>
    </child>
    <child groups='RW'>
      <object class="GtkToolButton">
        <property name="icon-name">list-remove</property>
        <property name="action-name">win.Remove</property>
        <property name="tooltip_text">%s</property>
        <property name="label" translatable="yes">_Delete</property>
        <property name="use-underline">True</property>
      </object>
      <packing>
        <property name="homogeneous">False</property>
      </packing>
    </child>
    <child groups='RW'>
      <object class="GtkToolButton">
        <property name="icon-name">gramps-merge</property>
        <property name="action-name">win.Merge</property>
        <property name="tooltip_text">%s</property>
        <property name="label" translatable="yes">_Merge...</property>
        <property name="use-underline">True</property>
      </object>
      <packing>
        <property name="homogeneous">False</property>
      </packing>
    </child>
    </placeholder>
"""
        % (ADD_MSG, ADD_SOURCE_MSG, ADD_CITATION_MSG, EDIT_MSG, DEL_MSG, MERGE_MSG),
        """
    <menu id="Popup">
      <section>
        <item>
          <attribute name="action">win.Back</attribute>
          <attribute name="label" translatable="yes">_Back</attribute>
        </item>
        <item>
          <attribute name="action">win.Forward</attribute>
          <attribute name="label" translatable="yes">Forward</attribute>
        </item>
      </section>
      <section id="PopUpTree">
        <item>
          <attribute name="action">win.OpenAllNodes</attribute>
          <attribute name="label" translatable="yes">"""
        """Expand all Nodes</attribute>
        </item>
        <item>
          <attribute name="action">win.CloseAllNodes</attribute>
          <attribute name="label" translatable="yes">"""
        """Collapse all Nodes</attribute>
        </item>
      </section>
       <section>
        <item>
          <attribute name="action">win.Add</attribute>
          <attribute name="label" translatable="yes">_Add...</attribute>
        </item>
        <item>
          <attribute name="action">win.AddCitation</attribute>
          <attribute name="label" translatable="yes">"""
        """Add citation...</attribute>
        </item>
        <item>
          <attribute name="action">win.Edit</attribute>
          <attribute name="label">%s</attribute>
        </item>
        <item>
          <attribute name="action">win.Remove</attribute>
          <attribute name="label" translatable="yes">_Delete</attribute>
        </item>
        <item>
          <attribute name="action">win.Merge</attribute>
          <attribute name="label" translatable="yes">_Merge...</attribute>
        </item>
      </section>
      <section>
        <placeholder id='QuickReport'>
        </placeholder>
      </section>
    </menu>
"""
        % _("_Edit...", "action"),  # to use sgettext()
    ]

    def add_source(self, *obj):
        """
        add_source: Add a new source (this is also available from the
                      source view)

        Create a new Source instance and call the EditSource editor with the
        new source.

        Called when the Add_source button is clicked.
        If the window already exists (WindowActiveError), we ignore it.
        This prevents the dialog from coming up twice on the same object.

        However, since the window is identified by the Source object, and
        we have just created a new one, it seems to be impossible for the
        window to already exist, so this is just an extra safety measure.
        """
        try:
            EditSource(self.dbstate, self.uistate, [], Source())
        except WindowActiveError:
            pass

    def add(self, *obj):
        """
        add:        Add a new citation and a new source (this can also be done
                      by source view add a source, then citation view add a new
                      citation to an existing source)

        Create a new Source instance and Citation instance and call the
        EditSource editor with the new source.

        Called when the Add button is clicked.
        If the window already exists (WindowActiveError), we ignore it.
        This prevents the dialog from coming up twice on the same object.

        However, since the window is identified by the Source object, and
        we have just created a new one, it seems to be impossible for the
        window to already exist, so this is just an extra safety measure.
        """
        try:
            EditCitation(self.dbstate, self.uistate, [], Citation(), Source())
        except WindowActiveError:
            pass

    def share(self, *obj):
        """
        share:      Add a new citation to an existing source (when a source is
                      selected)
        """
        for handle in self.selected_handles():
            # The handle will either be a Source handle or a Citation handle
            source, citation = self.get_source_or_citation(handle)
            if not source:
                source = self.dbstate.db.get_source_from_handle(
                    citation.get_reference_handle()
                )
            try:
                EditCitation(self.dbstate, self.uistate, [], Citation(), source)
            except WindowActiveError:
                from gramps.gui.dialog import WarningDialog

                WarningDialog(
                    _("Cannot share this reference"),
                    self.__blocked_text(),
                    parent=self.uistate.window,
                )

    def edit(self, *obj):
        """
        Edit either a Source or a Citation, depending on user selection
        """
        for handle in self.selected_handles():
            # The handle will either be a Source handle or a Citation handle
            source, citation = self.get_source_or_citation(handle)
            if citation:
                try:
                    EditCitation(self.dbstate, self.uistate, [], citation)
                except WindowActiveError:
                    pass
            else:
                try:
                    EditSource(self.dbstate, self.uistate, [], source)
                except WindowActiveError:
                    from gramps.gui.dialog import WarningDialog

                    WarningDialog(
                        _("Cannot share this reference"),
                        self.__blocked_text2(),
                        parent=self.uistate.window,
                    )

    def __blocked_text(self):
        """
        Return the common text used when citation cannot be edited
        """
        return _(
            "This citation cannot be created at this time. "
            "Either the associated Source object is already being "
            "edited, or another citation associated with the same "
            "source is being edited.\n\nTo edit this "
            "citation, you need to close the object."
        )

    def __blocked_text2(self):
        """
        Return the common text used when citation cannot be edited
        """
        return _(
            "This source cannot be edited at this time. "
            "Either the associated Source object is already being "
            "edited, or another citation associated with the same "
            "source is being edited.\n\nTo edit this "
            "source, you need to close the object."
        )

    def merge(self, *obj):
        """
        Merge the selected citations.
        """
        mlist = self.selected_handles()

        if len(mlist) != 2:
            msg = _("Cannot merge citations.")
            msg2 = _(
                "Exactly two citations must be selected to perform a "
                "merge. A second citation can be selected by holding "
                "down the control key while clicking on the desired "
                "citation."
            )
            ErrorDialog(msg, msg2, parent=self.uistate.window)
        else:
            source1, citation1 = self.get_source_or_citation(mlist[0])
            source2, citation2 = self.get_source_or_citation(mlist[1])

            if citation1 and citation2:
                if (
                    not citation1.get_reference_handle()
                    == citation2.get_reference_handle()
                ):
                    msg = _("Cannot merge citations.")
                    msg2 = _(
                        "The two selected citations must have the same "
                        "source to perform a merge. If you want to merge "
                        "these two citations, then you must merge the "
                        "sources first."
                    )
                    ErrorDialog(msg, msg2, parent=self.uistate.window)
                else:
                    MergeCitation(self.dbstate, self.uistate, [], mlist[0], mlist[1])
            elif source1 and source2:
                MergeSource(self.dbstate, self.uistate, [], mlist[0], mlist[1])
            else:
                msg = _("Cannot perform merge.")
                msg2 = _(
                    "Both objects must be of the same type, either "
                    "both must be sources, or both must be "
                    "citations."
                )
                ErrorDialog(msg, msg2, parent=self.uistate.window)
            self.object_build()

    def get_handle_from_gramps_id(self, gid):
        obj = self.dbstate.db.get_citation_from_gramps_id(gid)
        if obj:
            return obj.get_handle()
        else:
            return None

    def get_source_or_citation(self, handle, check=True):
        """
        Get one of source or citation from the passed handle.  There must be
        exactly one, or we will rain an exception.
        """
        try:
            source = self.dbstate.db.get_source_from_handle(handle)
        except HandleError:
            source = None
        try:
            citation = self.dbstate.db.get_citation_from_handle(handle)
        except HandleError:
            citation = None
        if check and ((not source and not citation) or (source and citation)):
            raise ValueError("selection must be either source or citation")
        return source, citation

    def tag_updated(self, handle_list):
        """
        Update tagged rows when a tag color changes.
        """
        all_links = set([])
        for tag_handle in handle_list:
            links = set(
                [
                    link[1]
                    for link in self.dbstate.db.find_backlink_handles(
                        tag_handle, include_classes="Citation"
                    )
                ]
            )
            all_links = all_links.union(links)
        self.row_update(list(all_links))

    def add_tag(self, transaction, handle, tag_handle):
        """
        Add the given tag to the given source or citation.
        """
        source, citation = self.get_source_or_citation(handle)
        if citation:
            citation.add_tag(tag_handle)
            self.dbstate.db.commit_citation(citation, transaction)
        else:
            source.add_tag(tag_handle)
            self.dbstate.db.commit_source(source, transaction)

    def remove_tag(self, transaction, handle, tag_handle):
        """
        Remove the given tag from the given source or citation.
        """
        source, citation = self.get_source_or_citation(handle)
        if citation:
            citation.remove_tag(tag_handle)
            self.dbstate.db.commit_citation(citation, transaction)
        else:
            source.remove_tag(tag_handle)
            self.dbstate.db.commit_source(source, transaction)

    def get_default_gramplets(self):
        """
        Define the default gramplets for the sidebar and bottombar.
        """
        return (
            ("Citation Filter",),
            ("Citation Gallery", "Citation Notes", "Citation Backlinks"),
        )

    def get_config_name(self):
        return __name__



#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2000-2007  Donald N. Allingham
# Copyright (C) 2008       Brian G. Matherly
# Copyright (C) 2010       Jakim Friant
# Copyright (C) 2010       Nick Hall
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

"""Tools/Database Processing/Edit Database Owner Information"""

# -------------------------------------------------------------------------
#
# gnome/gtk
#
# -------------------------------------------------------------------------
from gi.repository import Gtk

# -------------------------------------------------------------------------
#
# gramps modules
#
# -------------------------------------------------------------------------
from gramps.gen.const import URL_MANUAL_PAGE
from gramps.gen.config import config
from gramps.gen.utils.config import get_researcher
from gramps.gui.display import display_help
from gramps.gui.widgets import MonitoredEntry
from gramps.gui.managedwindow import ManagedWindow
from gramps.gui.plug import tool
from gramps.gen.const import GRAMPS_LOCALE as glocale

_ = glocale.translation.sgettext
from gramps.gui.glade import Glade
from gramps.gui.utils import is_right_click

# -------------------------------------------------------------------------
#
# Constants
#
# -------------------------------------------------------------------------
WIKI_HELP_PAGE = "%s_-_Tools" % URL_MANUAL_PAGE
WIKI_HELP_SEC = _("Edit_Database_Owner_Information", "manual")

# -------------------------------------------------------------------------
#
# constants
#
# -------------------------------------------------------------------------
config_keys = (
    "researcher.researcher-name",
    "researcher.researcher-addr",
    "researcher.researcher-locality",
    "researcher.researcher-city",
    "researcher.researcher-state",
    "researcher.researcher-country",
    "researcher.researcher-postal",
    "researcher.researcher-phone",
    "researcher.researcher-email",
)


# -------------------------------------------------------------------------
#
# OwnerEditor
#
# -------------------------------------------------------------------------
class OwnerEditor(tool.Tool, ManagedWindow):
    """
    Allow editing database owner information.

    Provides a possibility to directly verify and edit the owner data of the
    current database. It also allows copying data from/to the preferences.
    """

    def __init__(self, dbstate, user, options_class, name, callback=None):
        uistate = user.uistate
        ManagedWindow.__init__(self, uistate, [], self.__class__)
        tool.Tool.__init__(self, dbstate, options_class, name)

        self.display()

    def display(self):
        # get the main window from glade
        topDialog = Glade()

        # set gramps style title for the window
        window = topDialog.toplevel
        self.set_window(
            window, topDialog.get_object("title"), _("Database Owner Editor")
        )
        self.setup_configs("interface.ownereditor", 500, 400)

        # move help button to the left side
        action_area = topDialog.get_object("action_area")
        help_button = topDialog.get_object("help_button")
        action_area.set_child_secondary(help_button, True)

        # connect signals
        topDialog.connect_signals(
            {
                "on_ok_button_clicked": self.on_ok_button_clicked,
                "on_cancel_button_clicked": self.close,
                "on_help_button_clicked": self.on_help_button_clicked,
                "on_eventbox_button_press_event": self.on_button_press_event,
                "on_menu_activate": self.on_menu_activate,
            }
        )

        # fetch the popup menu
        self.menu = topDialog.get_object("popup_menu")
        self.track_ref_for_deletion("menu")

        # topDialog.connect_signals({"on_menu_activate": self.on_menu_activate})

        # get current db owner and attach it to the entries of the window
        self.owner = self.db.get_researcher()

        self.entries = []
        entry = [
            ("name", self.owner.set_name, self.owner.get_name),
            ("address", self.owner.set_address, self.owner.get_address),
            ("locality", self.owner.set_locality, self.owner.get_locality),
            ("city", self.owner.set_city, self.owner.get_city),
            ("state", self.owner.set_state, self.owner.get_state),
            ("country", self.owner.set_country, self.owner.get_country),
            ("zip", self.owner.set_postal_code, self.owner.get_postal_code),
            ("phone", self.owner.set_phone, self.owner.get_phone),
            ("email", self.owner.set_email, self.owner.get_email),
        ]

        for name, set_fn, get_fn in entry:
            self.entries.append(
                MonitoredEntry(
                    topDialog.get_object(name), set_fn, get_fn, self.db.readonly
                )
            )
        # ok, let's see what we've done
        self.show()

    def on_ok_button_clicked(self, obj):
        """Update the current db's owner information from editor"""
        self.db.set_researcher(self.owner)
        self.menu.destroy()
        self.close()

    def on_help_button_clicked(self, obj):
        """Display the relevant portion of Gramps manual"""
        display_help(webpage=WIKI_HELP_PAGE, section=WIKI_HELP_SEC)

    def on_button_press_event(self, obj, event):
        """Shows popup-menu for db <-> preferences copying"""
        if is_right_click(event):
            self.menu.popup_at_pointer(event)

    def build_menu_names(self, obj):
        return (_("Main window"), _("Edit database owner information"))

    def on_menu_activate(self, menuitem):
        """Copies the owner information from/to the preferences"""
        if menuitem.props.name == "copy_from_preferences_to_db":
            self.owner.set_from(get_researcher())
            for entry in self.entries:
                entry.update()

        elif menuitem.props.name == "copy_from_db_to_preferences":
            for i in range(len(config_keys)):
                config.set(config_keys[i], self.owner.get()[i])

    def clean_up(self):
        self.menu.destroy()


# -------------------------------------------------------------------------
#
# OwnerEditorOptions (None at the moment)
#
# -------------------------------------------------------------------------
class OwnerEditorOptions(tool.ToolOptions):
    """Defines options and provides handling interface."""

    def __init__(self, name, person_id=None):
        tool.ToolOptions.__init__(self, name, person_id)



# -*- coding: utf-8 -*-
#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2003-2005  Donald N. Allingham
# Copyright (C) 2008       Brian G. Matherly
# Copyright (C) 2009-2010  Andrew I Baznikin
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

# Written by Bernard Banko, inspired from rel_ru.py by Alex Roitman.
"""
Slovenian-specific definitions of relationships
"""

# -------------------------------------------------------------------------
#
# Gramps modules
#
# -------------------------------------------------------------------------

from gramps.gen.lib import Person
import gramps.gen.relationship

# -------------------------------------------------------------------------

_ancestors = ["", "starš", "stari starš", "prastari starš"]
_fathers = ["", "oče", "ded", "praded", "prapraded"]
_mothers = ["", "mati", "babica", "prababica", "praprababica"]
_descendants = ["", "otrok", "vnuk(inja)", "pravnuk(inja)", "prapravnuk(inja)"]
_sons = ["", "sin", "vnuk", "pravnuk", "prapravnuk"]
_daughters = ["", "hči", "vnukinja", "pravnukinja", "prapravnukinja"]
_maleCousins = ["", "brat", "bratranec", "mali bratranec"]
_femaleCousins = ["", "sestra", "sestrična", "mala sestrična"]
_someCousins = [
    "",
    "brat ali sestra",
    "bratranec ali sestrična",
    "mali bratranec ali mala sestrična",
]
_aunts = ["", "teta", "stara teta", "prateta", "praprateta"]
_uncles = ["", "stric", "stari stric", "prastric", "praprastric"]
_nieces = ["", "nečakinja", "pranečakinja", "prapranečakinja"]
_nephews = ["", "nečak", "pranečak", "prapranečak"]

# plural
_children = ["", "otroci", "vnuki", "pravnuki", "prapravnuki"]
_parents = ["", "starši", "stari starši", "prastarši", "praprastarši"]
_siblings = ["", "sorojenci", "strici", "stari strici", "prastrici", "praprastrici"]
_neph_niec = ["", "nečaki", "pranečaki", "prapranečaki"]


# -------------------------------------------------------------------------
#
#
#
# -------------------------------------------------------------------------
class RelationshipCalculator(gramps.gen.relationship.RelationshipCalculator):
    """
    RelationshipCalculator Class
    """

    def __init__(self):
        gramps.gen.relationship.RelationshipCalculator.__init__(self)

    def getAncestor(self, level):
        if level > len(_ancestors) - 1:
            return "%s-krat-pra-prednik" % (level - 2)
        else:
            return _ancestors[level]

    def getFather(self, level):
        if level > len(_fathers) - 1:
            return "%s-krat-pra-ded" % (level - 2)
        else:
            return _fathers[level]

    def getMother(self, level):
        if level > len(_mothers) - 1:
            return "%s-krat-pra-babica" % (level - 2)
        else:
            return _mothers[level]

    def getSon(self, level):
        if level > len(_sons) - 1:
            return "%s-krat-pra-vnuk" % (level - 2)
        else:
            return _sons[level]

    def getDaughter(self, level):
        if level > len(_daughters) - 1:
            return "%s-krat-pra-vnukinja" % (level - 2)
        else:
            return _daughters[level]

    def getDescendant(self, level):
        if level > len(_descendants) - 1:
            return "%s-krat-pra-vnuk(inja)" % (level - 2)
        else:
            return _descendants[level]

    def getMaleCousin(self, level):
        if level > len(_maleCousins) - 1:
            return "bratranec v %s. kolenu" % (level * 2)
        else:
            return _maleCousins[level]

    def getFemaleCousin(self, level):
        if level > len(_femaleCousins) - 1:
            return "sestrična v %s. kolenu" % (level * 2)
        else:
            return _femaleCousins[level]

    def getSomeCousin(self, level):
        if level > len(_someCousins) - 1:
            return "bratranec ali sestrična v %s. kolenu" % (level * 2)
        else:
            return _someCousins[level]

    def getSuffix(self, distance, level):
        # distance-level = 2Gb <=> Gb=1
        if distance - level == 2 or distance < 6:
            return ""
        else:
            return " v %s. kolenu" % (distance)

    def getAunt(self, distance, level):
        if distance == 5 and level == 1:
            return "mala teta"
        elif level > len(_aunts) - 1:
            return "%s-krat-pra-teta%s" % (level - 2, self.getSuffix(distance, level))
        else:
            return "%s%s" % (_aunts[level], self.getSuffix(distance, level))

    def getUncle(self, distance, level):
        if distance == 5 and level == 1:
            return "mali stric"
        elif level > len(_uncles) - 1:
            return "%s-krat-pra-stric%s" % (level - 2, self.getSuffix(distance, level))
        else:
            return "%s%s" % (_uncles[level], self.getSuffix(distance, level))

    def getNiece(self, distance, level):
        if distance == 5 and level == 1:
            return "mala nečakinja"
        elif level > len(_nieces) - 1:
            return "%s-krat-pra-nečakinja%s" % (
                level - 1,
                self.getSuffix(distance, level),
            )
        else:
            return "%s%s" % (_nieces[level], self.getSuffix(distance, level))

    def getNephew(self, distance, level):
        if distance == 5 and level == 1:
            return "mali nečak"
        elif level > len(_nephews) - 1:
            return "%s-krat-pra-nečak%s" % (level - 1, self.getSuffix(distance, level))
        else:
            return "%s%s" % (_nephews[level], self.getSuffix(distance, level))

    def get_single_relationship_string(
        self,
        Ga,
        Gb,
        gender_a,
        gender_b,
        reltocommon_a,
        reltocommon_b,
        only_birth=True,
        in_law_a=False,
        in_law_b=False,
    ):
        """
        Provide a string that describes the relationsip between a person, and
        another person. E.g. "grandparent" or "child".

        To be used as: 'person b is the grandparent of a', this will be in
        translation string:  'person b is the %(relation)s of a'

        Note that languages with gender should add 'the' inside the
        translation, so eg in french:  'person b est %(relation)s de a'
        where relation will be here: le grandparent

        Ga and Gb can be used to mathematically calculate the relationship.

        .. seealso::
            http://en.wikipedia.org/wiki/Cousin#Mathematical_definitions

        Some languages need to know the specific path to the common ancestor.
        Those languages should use reltocommon_a and reltocommon_b which is
        a string like 'mfmf'.

        The possible string codes are:

        =======================  ===========================================
        Code                     Description
        =======================  ===========================================
        REL_MOTHER               # going up to mother
        REL_FATHER               # going up to father
        REL_MOTHER_NOTBIRTH      # going up to mother, not birth relation
        REL_FATHER_NOTBIRTH      # going up to father, not birth relation
        REL_FAM_BIRTH            # going up to family (mother and father)
        REL_FAM_NONBIRTH         # going up to family, not birth relation
        REL_FAM_BIRTH_MOTH_ONLY  # going up to fam, only birth rel to mother
        REL_FAM_BIRTH_FATH_ONLY  # going up to fam, only birth rel to father
        =======================  ===========================================

        Prefix codes are stripped, so REL_FAM_INLAW_PREFIX is not present.
        If the relation starts with the inlaw of the person a, then 'in_law_a'
        is True, if it starts with the inlaw of person b, then 'in_law_b' is
        True.

        Also REL_SIBLING (# going sideways to sibling (no parents)) is not
        passed to this routine. The collapse_relations changes this to a
        family relation.

        Hence, calling routines should always strip REL_SIBLING and
        REL_FAM_INLAW_PREFIX before calling get_single_relationship_string()
        Note that only_birth=False, means that in the reltocommon one of the
        NOTBIRTH specifiers is present.

        The REL_FAM identifiers mean that the relation is not via a common
        ancestor, but via a common family (note that that is not possible for
        direct descendants or direct ancestors!). If the relation to one of the
        parents in that common family is by birth, then 'only_birth' is not
        set to False. The only_birth() method is normally used for this.

        :param Ga: The number of generations between the main person and the
                   common ancestor.
        :type Ga: int
        :param Gb: The number of generations between the other person and the
                   common ancestor.
        :type Gb: int
        :param gender_a: gender of person a
        :type gender_a: int gender
        :param gender_b: gender of person b
        :type gender_b: int gender
        :param reltocommon_a: relation path to common ancestor or common
                              Family for person a.
                              Note that length = Ga
        :type reltocommon_a: str
        :param reltocommon_b: relation path to common ancestor or common
                              Family for person b.
                              Note that length = Gb
        :type reltocommon_b: str
        :param in_law_a:  True if path to common ancestors is via the partner
                          of person a
        :type in_law_a: bool
        :param in_law_b: True if path to common ancestors is via the partner
                         of person b
        :type in_law_b: bool
        :param only_birth: True if relation between a and b is by birth only
                           False otherwise
        :type only_birth: bool
        :returns: A string describing the relationship between the two people
        :rtype: str

        .. note:: 1. the self.REL_SIBLING should not be passed to this routine,
                     so we should not check on it. All other self.
                  2. for better determination of siblings, use if Ga=1=Gb
                     get_sibling_relationship_string
        """
        if Gb == 0:
            if Ga == 0:
                rel_str = "ista oseba"
            elif gender_b == Person.MALE:
                rel_str = self.getFather(Ga)
            elif gender_b == Person.FEMALE:
                rel_str = self.getMother(Ga)
            else:
                rel_str = self.getAncestor(Ga)
        elif Ga == 0:
            if gender_b == Person.MALE:
                rel_str = self.getSon(Gb)
            elif gender_b == Person.FEMALE:
                rel_str = self.getDaughter(Gb)
            else:
                rel_str = self.getDescendant(Gb)
        elif Ga == Gb:
            if gender_b == Person.MALE:
                rel_str = self.getMaleCousin(Gb)
            elif gender_b == Person.FEMALE:
                rel_str = self.getFemaleCousin(Gb)
            else:
                rel_str = self.getSomeCousin(Gb)
        elif Ga > Gb:
            if gender_b == Person.FEMALE:
                rel_str = self.getAunt(Ga + Gb, Ga - Gb)
            else:
                rel_str = self.getUncle(
                    Ga + Gb, Ga - Gb
                )  # we'll use male for unknown sex
        else:  # Ga < Gb
            if gender_b == Person.FEMALE:
                rel_str = self.getNiece(Ga + Gb, Gb - Ga)
            else:
                rel_str = self.getNephew(
                    Ga + Gb, Gb - Ga
                )  # we'll use male for unknown sex
        return rel_str

    def get_sibling_relationship_string(
        self, sib_type, gender_a, gender_b, in_law_a=False, in_law_b=False
    ):
        """
        Determine the string giving the relation between two siblings of
        type sib_type.
        Eg: b is the brother of a
        Here 'brother' is the string we need to determine
        This method gives more details about siblings than
        get_single_relationship_string can do.

        .. warning:: DON'T TRANSLATE THIS PROCEDURE IF LOGIC IS EQUAL IN YOUR
                     LANGUAGE, AND SAME METHODS EXIST (get_uncle, get_aunt,
                     get_sibling)
        """
        gender = gender_b  # we don't need gender_a
        inlaw = in_law_a or in_law_b
        if sib_type == self.HALF_SIB_MOTHER or sib_type == self.HALF_SIB_FATHER:
            prefix = "pol"
        else:
            prefix = ""

        if sib_type < self.STEP_SIB:
            # ie. NORM_SIB or one of HALF_SIBs
            if not inlaw:
                if gender == Person.MALE:
                    rel_str = "%sbrat" % (prefix)
                elif gender == Person.FEMALE:
                    rel_str = "%ssestra" % (prefix)
                else:
                    rel_str = "%sbrat ali %ssestra" % (prefix, prefix)
            else:
                if gender == Person.MALE:
                    rel_str = "%ssvak" % (prefix)
                elif gender == Person.FEMALE:
                    rel_str = "%ssvakinja" % (prefix)
                else:
                    rel_str = "%ssvak ali %ssvakinja" % (prefix, prefix)
        else:
            rel_str = ""
        return rel_str

    def get_plural_relationship_string(
        self,
        Ga,
        Gb,
        reltocommon_a="",
        reltocommon_b="",
        only_birth=True,
        in_law_a=False,
        in_law_b=False,
    ):
        distance = Ga + Gb
        rel_str = "sorodniki v %s. kolenu" % (distance)
        if Ga == 0:
            # These are descendants
            if Gb < len(_children):
                rel_str = _children[Gb]
            else:
                rel_str = "%s-krat-pra-vnuki" % (Gb - 2)
        elif Gb == 0:
            # These are parents/grand parents
            if Ga < len(_parents):
                rel_str = _parents[Ga]
            else:
                rel_str = "%s-krat-pra-starši" % (Ga - 2)
        elif Gb == 1:
            # These are siblings/aunts/uncles
            if Ga < len(_siblings):
                rel_str = _siblings[Ga]
            else:
                rel_str = "%s-krat-pra-strici" % (Ga - 2)
        elif Ga == 1:
            # These are nieces/nephews
            if Gb < len(_neph_niec):
                rel_str = _neph_niec[Gb]
            else:
                rel_str = "%s-krat-pra-nečaki" % (Gb - 1)
        elif Ga == Gb:
            # These are cousins in the same generation
            if Ga == 2:
                rel_str = "bratranci"
            elif Ga == 3:
                rel_str = "mali bratranci"
            else:
                rel_str = "bratranci v %s. kolenu" % (distance)
        elif Ga > Gb:
            # These are cousins in different generations with the second person
            # being in a higher generation from the common ancestor than the
            # first person.
            level = Ga - Gb
            if distance == 5:
                rel_str = "mali strici"
            elif level < len(_siblings) - 1:
                # len-1 and level+1 to skip the siblings in uncles' levels
                rel_str = "%s v %s. kolenu" % (_siblings[level + 1], distance)
            else:
                rel_str = "%s-krat-pra-strici v %s. kolenu" % (level - 2, distance)
        else:  # Gb > Ga:
            # These are cousins in different generations with the second person
            # being in a lower generation from the common ancestor than the
            # first person.
            level = Gb - Ga
            if distance == 5:
                rel_str = "mali nečaki"
            elif level < len(_neph_niec):
                rel_str = "%s v %s. kolenu" % (_neph_niec[level], distance)
            else:
                rel_str = "%s-krat-pra-nečaki v %s. kolenu" % (level - 1, distance)
        if in_law_b == True:
            rel_str = "zakonci, ki jih imajo %s" % rel_str
        return rel_str


if __name__ == "__main__":
    """TRANSLATORS, copy this if statement at the bottom of your
    rel_xx.py module, and test your work with:
    python src/plugins/rel/rel_xx.py
    """
    from gramps.gen.relationship import test

    RC = RelationshipCalculator()
    test(RC, True)
#
# from PluginMgr import register_relcalc
#    register_relcalc(RelationshipCalculatorClass,["sl","sl_SI","sl-SI","slovenian","Slovenian","slovenščina"])



#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2000-2006  Donald N. Allingham
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

"""
UrlBase class for Gramps.
"""

# -------------------------------------------------------------------------
#
# Gramps modules
#
# -------------------------------------------------------------------------
from .url import Url
from .const import IDENTICAL, EQUAL


# -------------------------------------------------------------------------
#
# UrlBase classes
#
# -------------------------------------------------------------------------
class UrlBase:
    """
    Base class for url-aware objects.
    """

    def __init__(self, source=None):
        """
        Initialize an UrlBase.

        If the source is not None, then object is initialized from values of
        the source object.

        :param source: Object used to initialize the new object
        :type source: UrlBase
        """
        self.urls = list(map(Url, source.urls)) if source else []

    def serialize(self):
        """
        Convert the object to a serialized tuple of data
        """
        return [url.serialize() for url in self.urls]

    def unserialize(self, data):
        """
        Convert a serialized tuple of data to an object.
        """
        self.urls = [Url().unserialize(item) for item in data]

    def get_url_list(self):
        """
        Return the list of :class:`~.url.Url` instances associated with the
        object.

        :returns: List of :class:`~.url.Url` instances
        :rtype: list
        """
        return self.urls

    def set_url_list(self, url_list):
        """
        Set the list of :class:`~.url.Url` instances to passed the list.

        :param url_list: List of :class:`~.url.Url` instances
        :type url_list: list
        """
        self.urls = url_list

    def _merge_url_list(self, acquisition):
        """
        Merge the list of urls from acquisition with our own.

        :param acquisition: The url list of this object will be merged with
                            the current url list.
        :type acquisition: UrlBase
        """
        url_list = self.urls[:]
        for addendum in acquisition.get_url_list():
            for url in url_list:
                equi = url.is_equivalent(addendum)
                if equi == IDENTICAL:
                    break
                elif equi == EQUAL:
                    url.merge(addendum)
                    break
            else:
                self.urls.append(addendum)

    def add_url(self, url):
        """
        Add a :class:`~.url.Url` instance to the object's list of
        :class:`~.url.Url` instances.

        :param url: :class:`~.url.Url` instance to be added to the Person's
                    list of related web sites.
        :type url: :class:`~.url.Url`
        """
        self.urls.append(url)

    def remove_url(self, url):
        """
        Remove the specified :class:`~.url.Url` instance from the url list.

        If the instance does not exist in the list, the operation has no effect.

        :param url: :class:`~.url.Url` instance to remove from the list
        :type url: :class:`~.url.Url`

        :returns: True if the url was removed, False if it was not in the list.
        :rtype: bool
        """
        if url in self.urls:
            self.urls.remove(url)
            return True
        else:
            return False



#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2012       Doug Blank <doug.blank@gmail.com>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

"""
This package implements an object difference engine.
"""

import json

from ..db.utils import import_as_dict
from ..lib.serialize import to_json
from ..const import GRAMPS_LOCALE as glocale

_ = glocale.translation.gettext


def to_struct(obj):
    """
    Convert an object into a struct.
    """
    return json.loads(to_json(obj))


def diff_dates(json1, json2):
    """
    Compare two json date objects. Returns True if different.
    """
    if json1 == json2:  # if same, then Not Different
        return False  # else, they still might be Not Different
    elif isinstance(json1, dict) and isinstance(json2, dict):
        if json1["dateval"] == json2["dateval"] and json2["dateval"] != 0:
            return False
        elif json1["text"] == json2["text"]:
            return False
        else:
            return True
    else:
        return True


def diff_items(path, json1, json2):
    """
    Compare two json objects. Returns True if different.
    """
    if json1 == json2:
        return False
    elif isinstance(json1, list) and isinstance(json2, list):
        if len(json1) != len(json2):
            return True
        else:
            pos = 0
            for v1, v2 in zip(json1, json2):
                result = diff_items(path + ("[%d]" % pos), v1, v2)
                if result:
                    return True
                pos += 1
            return False
    elif isinstance(json1, dict) and isinstance(json2, dict):
        for key in json1.keys():
            if key == "change":
                continue  # don't care about time differences, only data changes
            elif key == "date":
                result = diff_dates(json1["date"], json2["date"])
                if result:
                    # print("different dates", path)
                    # print("   old:", json1["date"])
                    # print("   new:", json2["date"])
                    return True
            else:
                result = diff_items(path + "." + key, json1[key], json2[key])
                if result:
                    return True
        return False
    else:
        # print("different values", path)
        # print("   old:", json1)
        # print("   new:", json2)
        return True


def diff_dbs(db1, db2, user):
    """
    1. new objects => mark for insert
    2. deleted objects, no change locally after delete date => mark
       for deletion
    3. deleted objects, change locally => mark for user confirm for
       deletion
    4. updated objects => do a diff on differences, mark origin
       values as new data
    """
    missing_from_old = []
    missing_from_new = []
    diffs = []
    with user.progress(_("Family Tree Differences"), _("Searching..."), 10) as step:
        for item in [
            "Person",
            "Family",
            "Source",
            "Citation",
            "Event",
            "Media",
            "Place",
            "Repository",
            "Note",
            "Tag",
        ]:
            step()

            handles_func1 = db1.method("get_%s_handles", item)
            handles_func2 = db2.method("get_%s_handles", item)
            handle_func1 = db1.method("get_%s_from_handle", item)
            handle_func2 = db2.method("get_%s_from_handle", item)

            handles1 = sorted([handle for handle in handles_func1()])
            handles2 = sorted([handle for handle in handles_func2()])
            p1 = 0
            p2 = 0
            while p1 < len(handles1) and p2 < len(handles2):
                if handles1[p1] == handles2[p2]:  # in both
                    item1 = handle_func1(handles1[p1])
                    item2 = handle_func2(handles2[p2])
                    diff = diff_items(item, to_struct(item1), to_struct(item2))
                    if diff:
                        diffs += [(item, item1, item2)]
                    # else same!
                    p1 += 1
                    p2 += 1
                elif handles1[p1] < handles2[p2]:  # p1 is mssing in p2
                    item1 = handle_func1(handles1[p1])
                    missing_from_new += [(item, item1)]
                    p1 += 1
                elif handles1[p1] > handles2[p2]:  # p2 is mssing in p1
                    item2 = handle_func2(handles2[p2])
                    missing_from_old += [(item, item2)]
                    p2 += 1
            while p1 < len(handles1):
                item1 = handle_func1(handles1[p1])
                missing_from_new += [(item, item1)]
                p1 += 1
            while p2 < len(handles2):
                item2 = handle_func2(handles2[p2])
                missing_from_old += [(item, item2)]
                p2 += 1
    return diffs, missing_from_old, missing_from_new


def diff_db_to_file(old_db, filename, user):
    # First, get data as a InMemoryDB
    new_db = import_as_dict(filename, user, user)
    if new_db is not None:
        # Next get differences:
        diffs, m_old, m_new = diff_dbs(old_db, new_db, user)
        return diffs, m_old, m_new



# -*- coding: utf-8 -*-
#!/usr/bin/env python
#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2000-2007  Donald N. Allingham
# Copyright (C) 2007       Johan Gonqvist <johan.gronqvist@gmail.com>
# Copyright (C) 2007-2009  Gary Burton <gary.burton@zen.co.uk>
# Copyright (C) 2007-2009  Stephane Charette <stephanecharette@gmail.com>
# Copyright (C) 2008-2009  Brian G. Matherly
# Copyright (C) 2008       Jason M. Simanek <jason@bohemianalps.com>
# Copyright (C) 2008-2011  Rob G. Healey <robhealey1@gmail.com>
# Copyright (C) 2010       Doug Blank <doug.blank@gmail.com>
# Copyright (C) 2010       Jakim Friant
# Copyright (C) 2010-      Serge Noiraud
# Copyright (C) 2011       Tim G L Lyons
# Copyright (C) 2013       Benny Malengier
# Copyright (C) 2016       Allen Crider
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

"""
Narrative Web Page generator.

Classe:
    IntroductionPage
"""
# ------------------------------------------------
# python modules
# ------------------------------------------------
from decimal import getcontext
import logging

# ------------------------------------------------
# Gramps module
# ------------------------------------------------
from gramps.gen.const import GRAMPS_LOCALE as glocale
from gramps.plugins.lib.libhtml import Html

# ------------------------------------------------
# specific narrative web import
# ------------------------------------------------
from gramps.plugins.webreport.basepage import BasePage
from gramps.plugins.webreport.common import FULLCLEAR

_ = glocale.translation.sgettext
LOG = logging.getLogger(".NarrativeWeb")
getcontext().prec = 8


class IntroductionPage(BasePage):
    """
    This class is responsible for displaying information
    about the introduction page.
    """

    def __init__(self, report, the_lang, the_title):
        """
        @param: report    -- The instance of the main report class for
                             this report
        @param: the_lang  -- The lang to process
        @param: the_title -- The title page related to the language
        """
        BasePage.__init__(self, report, the_lang, the_title)
        ldatec = 0

        output_file, sio = self.report.create_file(report.intro_fname)
        result = self.write_header(self._("Introduction"))
        intropage, head, dummy_body, outerwrapper = result

        # begin Introduction division
        with Html("div", class_="content", id="Introduction") as section:
            outerwrapper += section

            introimg = self.add_image("introimg", head)
            if introimg is not None:
                section += introimg

            note_id = report.options["intronote"]
            ldatec = None
            if note_id:
                note = self.r_db.get_note_from_gramps_id(note_id)
                if note:
                    note_text = self.get_note_format(note, False)

                    # attach note
                    section += note_text

                    # last modification of this note
                    ldatec = note.get_change_time()

        # add clearline for proper styling
        # create footer section
        footer = self.write_footer(ldatec)
        outerwrapper += (FULLCLEAR, footer)

        # send page out for processing
        # and close the file
        self.xhtml_writer(intropage, output_file, sio, ldatec)



#!/usr/bin/python
# -*- coding: utf-8 -*-
#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2009 Gerald Britton <gerald.britton@gmail.com>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

"""
HTML operations.

This module exports the Html class

"""

# ------------------------------------------------------------------------
# Python modules
# ------------------------------------------------------------------------

# ------------------------------------------------------------------------
#
# Gramps modules
#
# ------------------------------------------------------------------------
from gramps.gen.const import GRAMPS_LOCALE as glocale

# ------------------------------------------------------------------------
#
# Constants
#
# ------------------------------------------------------------------------
__all__ = ["Html"]

# ------------------------------------------------------------------------
#
# XHTML DOCTYPE constants to be used in <!DOCTYPE ... > statements
#
# Reference: http://www.w3.org/QA/2002/04/valid-dtd-list.html
#
# ------------------------------------------------------------------------

_XHTML10_STRICT = (
    '"-//W3C//DTD XHTML 1.0 Strict//EN"\n'
    '\t"http://www.w3.org/TR/xhtml1/DTD/xhtml1-strict.dtd"'
)
_XTHML10_TRANS = (
    '"-//W3C//DTD XHTML 1.0 Transitional//EN"\n'
    '\t"http://www.w3.org/TR/xhtml1/DTD/xhtml1-transitional.dtd"'
)
_XHTML10_FRAME = (
    '"-//W3C//DTD XHTML 1.0 Frameset//EN"\n'
    '\t"http://www.w3.org/TR/xhtml1/DTD/xhtml1-frameset.dtd"'
)
_XHTML11 = (
    '"-//W3C//DTD XHTML 1.1//EN"\n' '\t"http://www.w3.org/TR/xhtml11/DTD/xhtml11.dtd"'
)
_XHTML10_BASIC = (
    '"-//W3C//DTD XHTML Basic 1.0//EN"\n'
    '\t"http://www.w3.org/TR/xhtml-basic/xhtml-basic10.dtd"'
)
_XHTML11_BASIC = (
    '"-//W3C//DTD XHTML Basic 1.1//EN"\n '
    '\t"http://www.w3.org/TR/xhtml-basic/xhtml-basic11.dtd"'
)
_HTML5 = ""

# ------------------------------------------------------------------------
#
# XML Namespace constant for use in <html xmlns=...> tags
#
# ------------------------------------------------------------------------

_XMLNS = "http://www.w3.org/1999/xhtml"

# ------------------------------------------------------------------------
#
# Local constants.
#
# ------------------------------------------------------------------------

# Set of html tags that do not use a complementary closing tag, but close with
# /> instead
_START_CLOSE = set(
    ["area", "base", "br", "frame", "hr", "img", "input", "link", "meta", "param"]
)

# ------------------------------------------------------------------------
#
# Html class.
#
# ------------------------------------------------------------------------


class Html(list):
    """
    HTML class: Manages a rooted tree of HTML objects
    """

    __slots__ = ["items", "indent", "inline", "close", "cms"]

    #
    @staticmethod
    def xmldecl(version=1.0, encoding="UTF-8", standalone="no"):
        """
        Build and return an XML declaration statement

        :type  version:    decimal number
        :param version:    version of XML to be used. Defaults to 1.0
        :type  encoding:   string
        :param encoding:   encoding method to be used. Defaults to "UTF-8"
        :type  standalone: string
        :param standalone: "yes" or "no".  Defaults to "no"
        """
        return "<?xml %s %s %s?>" % (
            'version="%s"' % version,
            'encoding="%s"' % encoding,
            'standalone="%s"' % standalone,
        )

    #
    @staticmethod
    def doctype(name="HTML", public="", external_id=_HTML5):
        """
        Build and return a DOCTYPE statement

        :type  name:        string
        :param name:        name of this DOCTYPE. Defaults to "html"
        :type  public:      string
        :param public:      class of this DOCTYPE. Defaults to 'PUBLIC
        :type  external_id: string
        :param external_id: external identifier of this DOCTYPE.
                            Defaults to XHTML 1.0 STRICT
        :type  args:        object
        :param args:        0 or more positional parameters to be added to
                            this DOCTYPE.
        """
        return (
            "<!DOCTYPE %s %s %s"
            % (
                name,
                public,
                external_id,
            )
        ).rstrip() + ">"

    #
    @staticmethod
    def html(xmlns=_XMLNS, lang="en", php_session=None, *args, **keywargs):
        """
        Build and return a properly-formated <html> object

        :type  xmlns: string
        :param xmlns: XML namespace string.
                      Default = 'http://www.w3.org/1999/xhtml'
        :type  lang:  string
        :param lang:  language to be used. Defaul = 'en'
        :rtype:       reference to new Html instance
        :php_session: If we need to have a php session start
        :returns:     reference to the newly-created Html instances
                      for <html> object
        """
        return Html(
            "html",
            indent=False,
            xmlns=xmlns,
            attr='xml:lang="%s" lang="%s"' % ((lang,) * 2),
            php=php_session,
            *args,
            **keywargs,
        )

    #
    @staticmethod
    def head(title=None, encoding="utf-8", html5=True, *args, **keywargs):
        """
        Build and return a properly-formated <head> object

        :type  title:    string or None
        :param title:    title for HTML page. Default=None. If None no
                         title tag is written
        :type  encoding: string
        :param encoding: encoding to be used. Default = 'utf-8'
        :param html5:    generate html5 syntax. Default = True. Set to False
                         if pre-html5 syntax required
        :rtype:          reference to new Html instance
        :returns:        reference to the newly-created Html instances
                         for <head> object
        """

        head = Html("head", *args, **keywargs)
        if title is not None:
            head += Html("title", title, inline=True, indent=True)
        if html5:
            head += Html("meta", charset=encoding, indent=True)
        else:
            meta1 = 'http-equiv="content-type" content="text/html;charset=%s"'
            meta2 = 'http-equiv="Content-Style-Type" content="text/css"'
            head += Html("meta", attr=meta1 % encoding, indent=True)
            head += Html("meta", attr=meta2, indent=True)
        return head

    #
    @staticmethod
    def page(
        title=None,
        encoding="utf-8",
        lang="en",
        html5=True,
        cms=False,
        php_session=None,
        *args,
        **keywargs,
    ):
        """
        This function prepares a new Html class based page and returns

        :type  title:    string
        :param title:    title for HTML page. Default=None
        :type  encoding: string
        :param encoding: encoding to be used. Default = 'utf-8'
        :type  lang:     string
        :param lang:     language to be used. Defaul = 'en'
        :param html5:    generate html5 syntax. Default = True. Set to False
                         if pre-html5 syntax required
        :rtype:          three object references
        :php_session:    the note to include before all html code
        :returns:        references to the newly-created Html instances for
                         page, head and body
        """
        page = Html.html(lang=lang, php_session=php_session, *args, **keywargs)
        if html5:
            page.addDOCTYPE(external_id=_HTML5)
        else:
            page.addXML(encoding=encoding)
            page.addDOCTYPE(external_id=_XHTML10_STRICT)
        #
        head = Html.head(
            title=title,
            encoding=encoding,
            lang=lang,
            html5=html5,
            indent=False,
            *args,
            **keywargs,
        )
        #
        if cms:
            body = Html("div", class_="body", indent=False, *args, **keywargs)
        else:
            body = Html("body", indent=False, *args, **keywargs)
        page += (head, body)
        return page, head, body

    #
    def __init__(self, tag="html", *args, **keywargs):
        """
        Class Constructor: Returns a new instance of the Html class

        :type  tag:      string
        :param tag:      The HTML tag. Default is 'html'
        :type  args:     optional positional parameters
        :param args:     0 more positional arguments to be inserted between
                         opening and closing HTML tags.
        :type  indent:   boolean or None
        :param indent:   True  ==> indent this object with respect to its
                                   parent
                         False ==> do not indent this object
                         None  ==> no indent for this object
                                   (use eg for pre tag)
                         Defaults to True
        :type  inline:   boolean
        :param inline:   True  ==> instructs the write() method to output
                                   this object and any child objects as a
                                   single string
                         False ==> output this object and its contents one
                                   string at a time
                         Defaults to False
        :type  close:    boolean or None
        :param close:    True  ==> this tag should be closed normally
                                   e.g. <tag>...</tag>
                         False ==> this tag should be automatically closed
                                   e.g. <tag />
                         None  ==> do not provide any closing for this tag
        :type  keywargs: optional keyword parameters
        :param keywargs: 0 or more keyword=argument pairs that should be
                         copied into the opening tag as keyword="argument"
                         attributes
        :rtype:          object reference
        :returns:        reference to the newly-created Html instance

        For full usage of the Html class with examples, please see the wiki
        page at: http://www.gramps-project.org/wiki/index.php?title=Libhtml
        """
        # Replace super(Html, self) with list
        # Issue with Python 2.6 and reload of plugin
        list.__init__(self, [])  # instantiate object
        attr = ""
        self.indent, self.close, self.inline = True, True, False
        #
        #       Handle keyword arguments passed to this constructor.
        #       Keywords that we process directly are handled.
        #       Keywords we don't recognize are saved for later
        #       addition to the opening tag as attributes.
        #
        phpnote = None
        for keyw, arg in sorted(keywargs.items()):
            if keyw in ["indent", "close", "inline"] and arg in [True, False, None]:
                setattr(self, keyw, arg)
            elif keyw == "attr":  # pass attributes along
                attr += " " + arg
            elif keyw == "php":  # php init session
                phpnote = arg
            elif keyw[-1] == "_":  # avoid Python conflicts
                attr += ' %s="%s"' % (keyw[:-1], arg)  # pass keyword arg along
            else:
                attr += ' %s="%s"' % (keyw, arg)  # pass keyword arg along
        #
        if tag[0] == "<":  # if caller provided preformatted tag?
            self[0:] = [tag]  #   add it in
            self.close = None  #   caller must close the tag
        else:
            if tag in _START_CLOSE:  # if tag in special list
                self.close = False  #   it needs no closing tag
            if phpnote:
                # We need to insert php code before the html tag
                # This is used to initiate a php session.
                htmlhead = phpnote + "<%s%s%s>"
            else:
                htmlhead = "<%s%s%s>"
            begin = htmlhead % (  # build opening tag with attributes
                tag,
                attr,
                ("" if self.close is not False else " /"),
            )
            #
            # Use slice syntax since we don't override slicing
            self[0:] = [begin] + list(args)  # add beginning tag
            if self.close:  # if need closing tab
                self[len(self) :] = ["</%s>" % tag]  # add it on the end

    #
    def __add(self, value):
        """
        Helper function for +, +=, operators and append() and extend()
        methods

        :type  value: object
        :param value: object to be added

        :rtype:       object reference
        :returns:     reference to object with new value added
        """
        if (
            isinstance(value, Html)
            or not hasattr(value, "__iter__")
            or isinstance(value, str)
        ):
            value = [value]
        index = len(self) - (1 if self.close else 0)
        self[index:index] = value
        return self

    #
    __iadd__ = __add__ = __add

    #
    def append(self, value):
        """
        Append a new value
        """
        self.__add(value)

    #
    extend = append

    #
    def replace(self, cur_value, value):
        """
        Replace current value with new value

        :type  cur_value: object
        :param cur_value: value of object to be replaced
        :type  value:     object
        :param value:     replacement value

        :rtype:           object reference
        :returns:         reference to object with new value added
        """
        self[self.index(cur_value)] = value

    #
    def __sub__(self, value):
        """
        Overload function for - and -= operators
        :type  value: object
        :param value: object to be removed

        :rtype:       object reference
        :returns:     reference to object with value removed
        """
        del self[self.index(value)]
        return self

    #
    __isub__ = remove = __sub__

    #
    def __str__(self):
        """
        Returns string representation

        :rtype:   string
        :returns: string representation of object
        """
        return "%s" * len(self) % tuple(self[:])

    #
    def __iter__(self):
        """
        Iterator function: returns a generator that performs an
        insertion-order tree traversal and yields each item found.
        """
        for item in self[:]:  # loop through all list elements
            if isinstance(item, Html):  # if nested list found
                for sub_item in item:  #     recurse
                    yield sub_item
            else:
                yield item

    #
    iterkeys = itervalues = iteritems = __iter__

    #
    def write(self, method=print, indent="\t", tabs=""):
        """
        Output function: performs an insertion-order tree traversal and
                         calls supplied method for each item found.

        :type  method:   function reference
        :param method:   function to call with each item found
        :type  indent:   string
        :param indenf:   string to use for indentation. Default = '\t' (tab)
        :type  tabs:     string
        :param tabs:     starting indentation
        """
        if self.indent is None:
            tabs = ""
        elif self.indent:
            tabs += indent
        if self.inline:  # if inline, write all list and
            method(str("%s%s" % (tabs, self)))  # nested list elements
        #
        else:
            for item in self[:]:  # else write one at a time
                if isinstance(item, Html):  # recurse if nested Html class
                    item.write(method=method, indent=indent, tabs=tabs)
                else:
                    method(str("%s%s" % (tabs, item)))  # else write the line

    #
    def addXML(self, version=1.0, encoding="UTF-8", standalone="no"):
        """
        Add an XML statement to the start of the list for this object

        :type  version:    decimal number
        :param version:    version of XML to be used. Defaults to 1.0
        :type  encoding:   string
        :param encoding:   encoding method to be used. Defaults to "UTF-8"
        :type  standalone: string
        :param standalone: "yes" or "no".  Defaults to "no"
        """
        xmldecl = Html.xmldecl(
            version=version, encoding=encoding, standalone=standalone
        )
        self[0:0] = [xmldecl]

    #
    def addDOCTYPE(self, name="html", public="PUBLIC", external_id=_HTML5, *args):
        """
        Add a DOCTYPE statement to the start of the list

        :type  name:        string
        :param name:        name of this DOCTYPE. Defaults to "html"
        :type  external_id: string
        :param external_id: external identifier of this DOCTYPE.
                            Defaults to XHTML 1.0 STRICT
        :type  args:        object
        :param args:        0 or more positional parameters to be added
                            to this DOCTYPE.
        """
        doctype = (
            "<!DOCTYPE %s %s %s%s"
            % (
                name,
                ("" if external_id == _HTML5 else public),
                external_id,
                " %s" * len(args) % args,
            )
        ).rstrip() + ">"
        # Note: DOCTYPE declaration must follow XML declaration

        # if len(self) and self[0][:6] == '<?xml ':
        if self[0][:6] == "<?xml ":
            self[1:1] = [doctype]
        else:
            self[0:0] = [doctype]

    #
    def __gettag(self):
        """
        Returns HTML tag for this object

        :rtype:   string
        :returns: HTML tag
        """
        return self[0].split()[0].strip("< >")

    #
    def __settag(self, newtag):
        """
        Sets a new HTML tag for this object

        :type  name: string
        :param name: new HTML tag
        """
        curtag = self.tag

        # Replace closing tag, if any

        if self[-1] == "</%s>" % curtag:
            self[-1] = "</%s>" % newtag

        # Replace opening tag

        self[0] = self[0].replace("<" + curtag, "<" + newtag)

    tag = property(__gettag, __settag)

    #
    def __getattr(self):
        """
        Returns HTML attributes for this object

        :rtype:   string
        :returns: HTML attributes
        """
        attr = self[0].strip("<!?/>").split(None, 1)
        return attr[1] if len(attr) > 1 else ""

    #
    def __setattr(self, value):
        """
        Sets new HTML attributes for this object

        :type  name: string
        :param name: new HTML attributes
        """
        beg = len(self.tag) + 1

        # See if self-closed or normal

        end = -2 if self.close is False else -1
        self[0] = self[0][:beg] + " " + value + self[0][end:]

    #
    def __delattr(self):
        """
        Removes HTML attributes for this object
        """
        self[0] = (
            "<"
            + self.tag
            + (  # Set correct closing delimiter(s)
                " />" if self.close is False else ">"
            )
        )

    #
    attr = property(__getattr, __setattr, __delattr)

    #
    def __getinside(self):
        """
        Returns list of items between opening and closing tags

        :rtype:   list
        :returns: list of items between opening and closing HTML tags
        """
        return self[1:-1]

    #
    def __setinside(self, value):
        """
        Sets new contents between opening and closing tags

        :type  name: list
        :param name: new HTML contents
        """
        if len(self) < 2:
            raise AttributeError("No closing tag. Cannot set inside value")
        if (
            isinstance(value, Html)
            or not hasattr(value, "__iter__")
            or isinstance(value, str)
        ):
            value = [value]
        self[1:-1] = value

    #
    def __delinside(self):
        """
        Removes contents between opening and closing tag
        """
        if len(self) > 2:
            self[:] = self[:1] + self[-1:]

    inside = property(__getinside, __setinside, __delinside)

    #
    def __enter__(self):
        return self

    #
    def __exit__(self, exc_type, exc_val, exc_tb):
        return exc_type is None


# ------------------------------------------------------------------------
#
# Functions
#
# ------------------------------------------------------------------------
def xml_lang():
    return glocale.lang[:5].replace("_", "-")


# -------------------------------------------
#
#           Unit tests
#
# -------------------------------------------


def htmltest():
    pass


if __name__ == "__main__":
    # from libhtmltest import htmltest
    # htmltest()
    pass



#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2000-2006  Donald N. Allingham
# Copyright (C) 2009-2011  Gary Burton
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

# -------------------------------------------------------------------------
#
# Python classes
#
# -------------------------------------------------------------------------
import os
import pickle
from urllib.parse import urlparse
from urllib.request import url2pathname

# -------------------------------------------------------------------------
#
# GTK libraries
#
# -------------------------------------------------------------------------
from gi.repository import Gdk
from gi.repository import GdkPixbuf
from gi.repository import Gtk
from gi.repository import Pango
from gi.repository import GObject
from gi.repository import GLib

# -------------------------------------------------------------------------
#
# Gramps classes
#
# -------------------------------------------------------------------------
from ...utils import is_right_click, open_file_with_default_application
from ...dbguielement import DbGUIElement
from ...selectors import SelectorFactory
from gramps.gen.lib import Media, MediaRef
from gramps.gen.db import DbTxn
from gramps.gen.utils.file import (
    media_path_full,
    media_path,
    relative_path,
    create_checksum,
)
from gramps.gen.utils.thumbnails import get_thumbnail_image
from gramps.gen.errors import WindowActiveError
from gramps.gen.mime import get_type, is_valid_type
from ...ddtargets import DdTargets
from .buttontab import ButtonTab
from gramps.gen.const import THUMBSCALE
from gramps.gen.const import GRAMPS_LOCALE as glocale

_ = glocale.translation.sgettext


# -------------------------------------------------------------------------
#
#
#
# -------------------------------------------------------------------------
def make_launcher(path, uistate):
    return lambda x: open_file_with_default_application(path, uistate)


# -------------------------------------------------------------------------
#
# GalleryTab
#
# -------------------------------------------------------------------------
class GalleryTab(ButtonTab, DbGUIElement):
    _DND_TYPE = DdTargets.MEDIAREF
    _DND_EXTRA = DdTargets.URI_LIST

    def __init__(self, dbstate, uistate, track, media_list, update=None):
        self.iconlist = Gtk.IconView()
        ButtonTab.__init__(
            self,
            dbstate,
            uistate,
            track,
            _("_Gallery"),
            True,
            move_buttons=ButtonTab.L_R,
        )
        DbGUIElement.__init__(self, dbstate.db)
        self.track_ref_for_deletion("iconlist")
        self.media_list = media_list
        self.callman.register_handles({"media": [mref.ref for mref in self.media_list]})
        self.update = update

        self._set_dnd()

        self.rebuild()
        self.show_all()

    def _connect_db_signals(self):
        """
        Implement base class DbGUIElement method
        """
        # note: media-rebuild closes the editors, so no need to connect to it
        self.callman.register_callbacks(
            {
                "media-delete": self.media_delete,  # delete a media we track
                "media-update": self.media_update,  # change a media we track
            }
        )
        self.callman.connect_all(keys=["media"])

    def double_click(self, obj, event):
        """
        Handle the button press event: double click or right click on iconlist.
        If the double click occurs, the Edit button handler is called.
        """
        if event.type == Gdk.EventType.DOUBLE_BUTTON_PRESS and event.button == 1:
            self.edit_button_clicked(obj)
            return True
        elif is_right_click(event):
            reflist = self.iconlist.get_selected_items()
            if len(reflist) == 1:
                path = reflist[0].get_indices()
                ref = self.media_list[path[0]]
                self.right_click(ref, event)
                return True
        return

    def right_click(self, obj, event):
        itemlist = [
            (True, _("_Add"), self.add_button_clicked),
            (True, _("Share"), self.share_button_clicked),
            (False, _("_Edit"), self.edit_button_clicked),
            (True, _("_Remove"), self.del_button_clicked),
        ]

        self.menu = Gtk.Menu()
        self.menu.set_reserve_toggle_size(False)

        ref_obj = self.dbstate.db.get_media_from_handle(obj.ref)
        media_path = media_path_full(self.dbstate.db, ref_obj.get_path())
        if media_path:
            # Translators: _View means "to look at this"
            item = Gtk.MenuItem.new_with_mnemonic(_("_View", "verb:look at this"))
            item.connect("activate", make_launcher(media_path, self.uistate))
            item.show()
            self.menu.append(item)
            mfolder, mfile = os.path.split(media_path)
            item = Gtk.MenuItem.new_with_mnemonic(_("Open Containing _Folder"))
            item.connect("activate", make_launcher(mfolder, self.uistate))
            item.show()
            self.menu.append(item)
            item = Gtk.SeparatorMenuItem()
            item.show()
            self.menu.append(item)

            item = Gtk.MenuItem.new_with_mnemonic(_("_Make Active Media"))
            item.connect(
                "activate", lambda obj: self.uistate.set_active(ref_obj.handle, "Media")
            )
            item.show()
            self.menu.append(item)
            item = Gtk.SeparatorMenuItem()
            item.show()
            self.menu.append(item)

        for needs_write_access, title, func in itemlist:
            item = Gtk.MenuItem.new_with_mnemonic(title)
            item.connect("activate", func)
            if needs_write_access and self.dbstate.db.readonly:
                item.set_sensitive(False)
            item.show()
            self.menu.append(item)
        self.menu.popup_at_pointer(event)

    def get_icon_name(self):
        return "gramps-media"

    def is_empty(self):
        return len(self.media_list) == 0

    def _build_icon_model(self):
        self.iconmodel = Gtk.ListStore(GdkPixbuf.Pixbuf, GObject.TYPE_STRING, object)
        self.track_ref_for_deletion("iconmodel")

    def _connect_icon_model(self):
        self.iconlist.set_model(self.iconmodel)
        self.iconmodel.connect_after("row-inserted", self._update_internal_list)
        self.iconmodel.connect_after("row-deleted", self._update_internal_list)

    def build_interface(self):
        """Setup the GUI.

        It includes an IconView placed inside of a ScrolledWindow.

        """
        # create the model used with the icon view
        self._build_icon_model()

        # pixels to pad the image
        padding = 6

        # build the icon view
        self.iconlist.set_pixbuf_column(0)
        self.iconlist.set_item_width(int(THUMBSCALE) + padding * 2)
        # set custom text cell renderer for better control
        text_renderer = Gtk.CellRendererText()
        text_renderer.set_property("wrap-mode", Pango.WrapMode.WORD_CHAR)
        text_renderer.set_property("wrap-width", THUMBSCALE)
        text_renderer.set_property("alignment", Pango.Alignment.CENTER)
        self.iconlist.pack_end(text_renderer, True)
        self.iconlist.add_attribute(text_renderer, "text", 1)

        # set basic properties of the icon view
        self.iconlist.set_margin(padding)
        self.iconlist.set_column_spacing(padding)
        self.iconlist.set_reorderable(True)
        self.iconlist.set_selection_mode(Gtk.SelectionMode.SINGLE)

        # connect the signals
        self.iconlist.connect("selection-changed", self._selection_changed)
        self.iconlist.connect("button_press_event", self.double_click)
        self.iconlist.connect("key_press_event", self.key_pressed)
        self._connect_icon_model()

        # create the scrolled window
        scroll = Gtk.ScrolledWindow()
        scroll.set_policy(Gtk.PolicyType.AUTOMATIC, Gtk.PolicyType.AUTOMATIC)

        # put everything together
        scroll.add(self.iconlist)
        self.pack_end(scroll, True, True, 0)

    def _update_internal_list(self, *obj):
        newlist = []
        node = self.iconmodel.get_iter_first()
        while node is not None:
            newlist.append(self.iconmodel.get_value(node, 2))
            node = self.iconmodel.iter_next(node)
        for i in self.media_list:
            self.media_list.pop()
        for i in newlist:
            if i:
                self.media_list.append(i)

        if self.update:
            self.update()
        self.changed = True

    def get_data(self):
        return self.media_list

    def rebuild(self):
        self._build_icon_model()
        for ref in self.media_list:
            handle = ref.get_reference_handle()
            obj = self.dbstate.db.get_media_from_handle(handle)
            if obj is None:
                # notify user of error
                from ...dialog import RunDatabaseRepair

                RunDatabaseRepair(
                    _("Non existing media found in the Gallery"),
                    parent=self.uistate.window,
                )
            else:
                pixbuf = get_thumbnail_image(
                    media_path_full(self.dbstate.db, obj.get_path()),
                    obj.get_mime_type(),
                    ref.get_rectangle(),
                )
                self.iconmodel.append(row=[pixbuf, obj.get_description(), ref])
        self._connect_icon_model()
        self._set_label()
        self._selection_changed()
        if self.update:
            self.update()

    def get_selected(self):
        node = self.iconlist.get_selected_items()
        if len(node) > 0:
            path = node[0].get_indices()
            return self.media_list[path[0]]
        return None

    def add_button_clicked(self, obj):
        try:
            from .. import EditMediaRef

            EditMediaRef(
                self.dbstate,
                self.uistate,
                self.track,
                Media(),
                MediaRef(),
                self.add_callback,
            )
        except WindowActiveError:
            pass

    def add_callback(self, media_ref, media):
        media_ref.ref = media.handle
        data = self.get_data()
        data.append(media_ref)
        self.callman.register_handles({"media": [media.handle]})
        self.changed = True
        self.rebuild()
        model = self.iconlist.get_model()
        if model:
            itr_last = model.iter_nth_child(None, len(data) - 1)
            if itr_last:
                path = model.get_path(itr_last)
                GLib.idle_add(self.iconlist.scroll_to_path, path, False, 0.0, 0.0)

    def __blocked_text(self):
        """
        Return the common text used when mediaref cannot be edited
        """
        return _(
            "This media reference cannot be edited at this time. "
            "Either the associated media object is already being "
            "edited or another media reference that is associated with "
            "the same media object is being edited.\n\nTo edit this "
            "media reference, you need to close the media object."
        )

    def share_button_clicked(self, obj):
        """
        Function called when the Share button is clicked.

        This function should be overridden by the derived class.

        """
        SelectObject = SelectorFactory("Media")

        sel = SelectObject(self.dbstate, self.uistate, self.track)
        src = sel.run()
        if src:
            sref = MediaRef()
            try:
                from .. import EditMediaRef

                EditMediaRef(
                    self.dbstate, self.uistate, self.track, src, sref, self.add_callback
                )
            except WindowActiveError:
                from ...dialog import WarningDialog

                WarningDialog(
                    _("Cannot share this reference"),
                    self.__blocked_text(),
                    parent=self.uistate.window,
                )

    def del_button_clicked(self, obj):
        ref = self.get_selected()
        if ref:
            self.media_list.remove(ref)
            self.rebuild()

    def edit_button_clicked(self, obj):
        ref = self.get_selected()
        if ref:
            obj = self.dbstate.db.get_media_from_handle(ref.get_reference_handle())
            try:
                from .. import EditMediaRef

                EditMediaRef(
                    self.dbstate, self.uistate, self.track, obj, ref, self.edit_callback
                )
            except WindowActiveError:
                from ...dialog import WarningDialog

                WarningDialog(
                    _("Cannot edit this reference"),
                    self.__blocked_text(),
                    parent=self.uistate.window,
                )

    def edit_callback(self, media_ref, media):
        """
        Rebuild the gallery after a media reference is edited in case the
        image rectangle has changed.
        """
        self.rebuild()

    def media_delete(self, del_media_handle_list):
        """
        Outside of this tab media objects have been deleted. Check if tab
        and object must be changed.
        Note: delete of object will cause reference on database to be removed,
              so this method need not do this
        """
        rebuild = False
        ref_handles = [x.ref for x in self.media_list]
        for handle in del_media_handle_list:
            while 1:
                pos = None
                try:
                    pos = ref_handles.index(handle)
                except ValueError:
                    break

                if pos is not None:
                    # oeps, we need to remove this reference, and rebuild tab
                    del self.media_list[pos]
                    del ref_handles[pos]
                    rebuild = True
        if rebuild:
            self.rebuild()

    def media_update(self, upd_media_handle_list):
        """
        Outside of this tab media objects have been changed. Check if tab
        and object must be changed.
        """
        ref_handles = [x.ref for x in self.media_list]
        for handle in upd_media_handle_list:
            if handle in ref_handles:
                self.rebuild()
                break

    def up_button_clicked(self, obj):
        """Deal with up button"""
        ref = self.get_selected()
        if ref:
            pos = self.find_index(ref)
            if pos > 0:
                self._move_up(pos, ref)

    def down_button_clicked(self, obj):
        """Deal with down button"""
        ref = self.get_selected()
        if ref:
            pos = self.find_index(ref)
            if pos >= 0 and pos < len(self.get_data()) - 1:
                self._move_down(pos, ref)

    def _move_up(self, row_from, obj):
        """
        Move the item a position up in the EmbeddedList.
        Eg: 0,1,2,3 needs to become 0,2,1,3, here row_from = 2
        """
        dlist = self.get_data()
        del dlist[row_from]
        dlist.insert(row_from - 1, obj)
        self.changed = True
        self.rebuild()
        # select the row
        path = Gtk.TreePath.new_from_string(str(row_from - 1))
        self.iconlist.select_path(path)

    def _move_down(self, row_from, obj):
        """
        Move the item a position down in the EmbeddedList.
        Eg: 0,1,2,3 needs to become 0,2,1,3, here row_from = 1
        """
        dlist = self.get_data()
        del dlist[row_from]
        dlist.insert(row_from + 1, obj)
        self.changed = True
        self.rebuild()
        # select the row
        path = Gtk.TreePath.new_from_string(str(row_from + 1))
        self.iconlist.select_path(path)

    def _set_dnd(self):
        """
        Set up drag-n-drop. The source and destination are set by calling .target()
        on the _DND_TYPE. Obviously, this means that there must be a _DND_TYPE
        variable defined that points to an entry in DdTargets.
        """

        dnd_types = [
            self._DND_TYPE.target(),
            self._DND_EXTRA.target(),
            DdTargets.MEDIAOBJ.target(),
        ]

        self.iconlist.enable_model_drag_dest(
            dnd_types, Gdk.DragAction.MOVE | Gdk.DragAction.COPY
        )
        self.iconlist.enable_model_drag_source(
            Gdk.ModifierType.BUTTON1_MASK,
            [self._DND_TYPE.target()],
            Gdk.DragAction.COPY,
        )
        self.iconlist.connect("drag_data_get", self.drag_data_get)
        if not self.dbstate.db.readonly:
            self.iconlist.connect("drag_data_received", self.drag_data_received)

    def drag_data_get(self, widget, context, sel_data, info, time):
        """
        Provide the drag_data_get function, which passes a tuple consisting of:

           1) Drag type defined by the .drag_type field specified by the value
              assigned to _DND_TYPE
           2) The id value of this object, used for the purpose of determining
              the source of the object. If the source of the object is the same
              as the object, we are doing a reorder instead of a normal drag
              and drop
           3) Pickled data. The pickled version of the selected object
           4) Source row. Used for a reorder to determine the original position
              of the object
        """

        # get the selected object, returning if not is defined

        try:
            reflist = self.iconlist.get_selected_items()
            path = reflist[0].get_indices()
            obj = self.media_list[path[0]]

            if not obj:
                return

            # pickle the data, and build the tuple to be passed
            value = (self._DND_TYPE.drag_type, id(self), obj, self.find_index(obj))
            data = pickle.dumps(value)

            # pass as a string (8 bits)
            sel_data.set(self._DND_TYPE.atom_drag_type, 8, data)
        except IndexError:
            return

    def drag_data_received(self, widget, context, x, y, sel_data, info, time):
        """
        Handle the standard gtk interface for drag_data_received.

        If the selection data is define, extract the value from sel_data.data,
        and decide if this is a move or a reorder.
        """
        if sel_data and sel_data.get_data():
            sel_list = []
            try:
                rets = pickle.loads(sel_data.get_data())

                # single dnd item
                if isinstance(rets, tuple):
                    sel_list.append(rets)

                # multiple dnd items
                elif isinstance(rets, list):
                    for ret in rets:
                        sel_list.append(pickle.loads(ret))

                for sel in sel_list:
                    (mytype, selfid, obj, row_from) = sel

                    # make sure this is the correct DND type for this object
                    if mytype == self._DND_TYPE.drag_type:
                        # determine the destination row
                        data = self.iconlist.get_dest_item_at_pos(x, y)
                        if data:
                            (path, pos) = data
                            row = path.get_indices()[0]
                            if pos == Gtk.IconViewDropPosition.DROP_LEFT:
                                row = max(row, 0)
                            elif pos == Gtk.IconViewDropPosition.DROP_RIGHT:
                                row = min(row, len(self.get_data()))
                            elif pos == Gtk.IconViewDropPosition.DROP_INTO:
                                row = min(row + 1, len(self.get_data()))
                        else:
                            row = len(self.get_data())

                        # if the is same object, we have a move, otherwise,
                        # it is a standard drag-n-drop

                        if id(self) == selfid:
                            self._move(row_from, row, obj)
                        else:
                            self._handle_drag(row, obj)
                        self.rebuild()
                    elif mytype == DdTargets.MEDIAOBJ.drag_type:
                        oref = MediaRef()
                        oref.set_reference_handle(obj)
                        self.get_data().append(oref)
                        self.changed = True
                        self.rebuild()
                    elif self._DND_EXTRA and mytype == self._DND_EXTRA.drag_type:
                        self.handle_extra_type(mytype, obj)
            except pickle.UnpicklingError:
                files = sel_data.get_uris()
                for file in files:
                    protocol, site, mfile, j, k, l = urlparse(file)
                    if protocol == "file":
                        name = url2pathname(mfile)
                        mime = get_type(name)
                        if not is_valid_type(mime):
                            return
                        photo = Media()
                        self.uistate.set_busy_cursor(True)
                        photo.set_checksum(create_checksum(name))
                        self.uistate.set_busy_cursor(False)
                        base_dir = str(media_path(self.dbstate.db))
                        if os.path.exists(base_dir):
                            name = relative_path(name, base_dir)
                        photo.set_path(name)
                        photo.set_mime_type(mime)
                        basename = os.path.basename(name)
                        (root, ext) = os.path.splitext(basename)
                        photo.set_description(root)
                        with DbTxn(_("Drag Media Object"), self.dbstate.db) as trans:
                            self.dbstate.db.add_media(photo, trans)
                            oref = MediaRef()
                            oref.set_reference_handle(photo.get_handle())
                            self.get_data().append(oref)
                            self.changed = True
                    self.rebuild()

    def handle_extra_type(self, objtype, obj):
        pass

    def _handle_drag(self, row, obj):
        self.get_data().insert(row, obj)
        self.changed = True

    def _move(self, row_from, row_to, obj):
        dlist = self.get_data()
        if row_from < row_to:
            dlist.insert(row_to, obj)
            del dlist[row_from]
        else:
            del dlist[row_from]
            dlist.insert(row_to, obj)
        self.changed = True

    def find_index(self, obj):
        """
        returns the index of the object within the associated data
        """
        return self.get_data().index(obj)

    def clean_up(self):
        super(ButtonTab, self).clean_up()
